#
# equalize.py
# Francesco Conti <fconti@iis.ee.ethz.ch>
#
# Copyright (C) 2018-2020 ETH Zurich
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from nemo.precision import Precision
from nemo.quant.pact import *
from nemo.graph import DeployGraph
from torch.nn.modules.utils import _single,_pair
from collections import OrderedDict
import types
import logging
import numpy as np
import copy
import math
import torchvision.models
import re
from nemo.transf.common import *
from sklearn import linear_model

# Part of the procedure necessary for DFQ as described here https://arxiv.org/pdf/1906.04721.pdf            
def _equalize_weights_dfq_pact(self, equalize_dict={}, act_dict={}, verbose=False, cost_eps=1e-3, max_iter=1000, reset_alpha=True):
    r"""This function implements the cross-layer weight-range equalization procedure proposed in 
    the Data-Free Quantization paper by Qualcomm (https://arxiv.org/pdf/1906.04721.pdf).
    It should be used only after batch-normalization layers have been folded into convolution
    by means of the `fold_bn` or `fold_bn_withinv` methods.
    
    :param equalize_dict: a dictionary of layer names, with the key being a Linear and the value the next Linear layer.
    :type  equalize_dict: `dict` or `collections.OrderedDict`
    :param act_dict: a dictionary of layer names, with the key being a Linear and the value the next Act layer. If empty, activation alpha scaling is not performed unless `equalize_dict` is also empty.
    :type  act_dict: `dict` or `collections.OrderedDict`
    :param verbose: if True, prints more information.
    :type  verbose: bool
    :param cost_eps: equalization will iterate until the cost is less than this threshold or the number of iterations is greater than `max_iter`.
    :type  cost_eps: float
    :param max_iter: maximum number of iterations.
    :type  max_iter: int
    :param reset_alpha: if True, reset the clipping parameters of weights (default True).
    :type  reset_alpha: bool

    """

    if not equalize_dict:
        equalize_dict, act_dict = get_equalize_dict_from_supernodes(self)

    module_dict = {}
    for n,m in self.named_modules():
        if (m.__class__.__name__ == "PACT_Conv2d" or \
            m.__class__.__name__ == "PACT_Conv1d" or \
            m.__class__.__name__ == "PACT_Linear" or \
            m.__class__.__name__ == "BatchNorm2d" or \
            m.__class__.__name__ == "BatchNorm1d" or \
            m.__class__.__name__ == "PACT_Act"):
            module_dict[n] = m
    it = 0
    cost = 1e10
    while cost > cost_eps and it < max_iter:
        cost = 0.0
        for n_before in equalize_dict.keys():
            n_after = equalize_dict[n_before]
            m_before = module_dict[n_before]
            m_after  = module_dict[n_after]
            range_before = weight_range(m_before, 0)
            range_after  = weight_range(m_after, 1)
            old_prec_before_mean = (weight_range(m_before, 0).abs() / (m_before.weight.max() - m_before.weight.min()).abs()).sum().item()
            old_prec_after_mean  = (weight_range(m_after, 0).abs()  / (m_after.weight.max()  - m_after.weight.min()).abs()).sum().item()

            # this happens when the two layers are across a Flatten operation
            flatten_flag = False
            if range_after.shape[0] != range_before.shape[0]:
                flatten_flag = True
                range_after = range_after.reshape((range_before.shape[0], -1))
                flatten_dim = range_after.shape[1]
                range_after = range_after.max(1)[0]

            s = torch.sqrt(range_after/range_before)
            m_before.weight.data[:] = m_before.weight.data[:] * reshape_before(m_before, s)
            if act_dict:
                # per-layer: has to use s max!
                module_dict[act_dict[n_before]].alpha.data[:] *= s.max()
            try:
                m_before.bias.data[:] = m_before.bias.data[:] * s
            except AttributeError:
                pass

            if flatten_flag:
                s = torch.cat(flatten_dim*(s.unsqueeze(1),),1).flatten()

            m_after.weight.data[:]  = m_after.weight.data[:] / reshape_after(m_after, s)
            new_prec_before_mean = (weight_range(m_before, 0).abs() / (m_before.weight.max() - m_before.weight.min()).abs()).sum().item()
            new_prec_after_mean  = (weight_range(m_after, 0).abs()  / (m_after.weight.max()  - m_after.weight.min()).abs()).sum().item()
            cost += np.abs(new_prec_before_mean*new_prec_after_mean - old_prec_before_mean*old_prec_after_mean)
        it += 1
        if verbose:
            logging.info("[DFQ Equalization] cost=%.5f" % cost)
    logging.info("[DFQ Equalization] terminated after %d iterations" % it)
    if reset_alpha:
        self.reset_alpha_weights()

def _equalize_weights_unfolding_pact(self, bn_dict={}, verbose=False, eps=None):
    r"""Performs in-layer equalization by unfolding of convolution parameters
    into batch-normalization parameters.
    
    :param bn_dict: a dictionary of layer names, with the key being the source (linear) and the value the target (batch-norm).
    :type  bn_dict: `dict` or `collections.OrderedDict`
    :param verbose: if True, prints more information.
    :type  verbose: bool
    :param eps: if not None (the default), overrides numerical `eps` used within batch-norm layer.
    :type  eps: float

    """

    if not bn_dict:
        bn_dict = get_bn_dict_from_supernodes(self)

    module_dict = {}
    for n,m in self.named_modules():
        if (m.__class__.__name__ == "PACT_Conv2d" or \
            m.__class__.__name__ == "PACT_Conv1d" or \
            m.__class__.__name__ == "PACT_Linear" or \
            m.__class__.__name__ == "BatchNorm2d" or \
            m.__class__.__name__ == "BatchNorm1d" ):
            module_dict[n] = m
    for n_before in bn_dict.keys():
        n_after = bn_dict[n_before]
        m_before = module_dict[n_before]
        m_after  = module_dict[n_after]
        if eps is None:
            eps = m_after.eps
        range_before = weight_range(m_before, 0)
        if verbose:
            logging.info("[Equalization by Unfolding] %s: wrange_min=%.5f wrange_max=%.5f" % (n_before, range_before.min().item(), range_before.max().item()))
        m_before.weight.data[:] = m_before.weight.data[:] / reshape_before(m_before, range_before)
        try:
            m_before.bias.data[:] = m_before.bias.data[:] / range_before
        except AttributeError:
            pass
        m_after.running_mean.data[:] = m_after.running_mean.data[:] / range_before
        m_after.weight.data[:] = m_after.weight.data[:] * reshape_after(m_after, range_before)
        if verbose:
            logging.info("[Equalization by Unfolding] %s: wrange_min=%.5f wrange_max=%.5f" % (n_before, weight_range(m_before, 0).min().item(), weight_range(m_before, 0).max().item()))

def _equalize_weights_lsq_pact(self, bn_dict={}, verbose=False, eps=None):
    r"""Performs in-layer equalization by unfolding of convolution parameters
    into batch-normalization parameters.
    
    :param bn_dict: a dictionary of layer names, with the key being the source (linear) and the value the target (batch-norm).
    :type  bn_dict: `dict` or `collections.OrderedDict`
    :param verbose: if True, prints more information.
    :type  verbose: bool
    :param eps: if not None (the default), overrides numerical `eps` used within batch-norm layer.
    :type  eps: float

    """

    if not bn_dict:
        bn_dict = get_bn_dict_from_supernodes(self)

    module_dict = {}
    for n,m in self.named_modules():
        if (m.__class__.__name__ == "PACT_Conv2d" or \
            m.__class__.__name__ == "PACT_Conv1d" or \
            m.__class__.__name__ == "PACT_Linear" or \
            m.__class__.__name__ == "BatchNorm2d" or \
            m.__class__.__name__ == "BatchNorm1d" ):
            module_dict[n] = m
    for n_before in bn_dict.keys():
        n_after = bn_dict[n_before]
        m_before = module_dict[n_before]
        m_after  = module_dict[n_after]
        if eps is None:
            eps = m_after.eps
        min_before = weight_min(m_before, 0).cpu().detach().numpy()
        max_before = weight_max(m_before, 0).cpu().detach().numpy()
        if verbose:
            logging.info("[Equalization by Least Squares] %s: wrange_min=%.5f wrange_max=%.5f" % (n_before, weight_range(m_before, 0).min().item(), weight_range(m_before, 0).max().item()))
        X = np.vstack((min_before, max_before))
        y = np.asarray((-1,1))
        coeff = torch.zeros(len(min_before), device=m_before.weight.device)
        regr = linear_model.LinearRegression(fit_intercept=False)
        for i in range(len(min_before)):
            regr.fit(X[:,i].reshape((-1,1)), y)
            coeff[i] = torch.as_tensor(regr.coef_[0], device=m_before.weight.device)
        coeff = 1./coeff
        m_before.weight.data[:] = m_before.weight.data[:] / reshape_before(m_before, coeff)
        try:
            m_before.bias.data[:] = m_before.bias.data[:] / coeff
        except AttributeError:
            pass
        m_after.running_mean.data[:] = m_after.running_mean.data[:] / coeff
        m_after.weight.data[:] = m_after.weight.data[:] * reshape_after(m_after, coeff)
        if verbose:
            logging.info("[Equalization by Least Squares] %s: wrange_min=%.5f wrange_max=%.5f" % (n_before, weight_range(m_before, 0).min().item(), weight_range(m_before, 0).max().item()))




#
# graph.py
# Francesco Conti <fconti@iis.ee.ethz.ch>
#
# Copyright (C) 2018-2020 ETH Zurich
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import torch.onnx.utils
from nemo.precision import Precision
from nemo.quant.pact import *
from torch.nn.modules.utils import _single,_pair
from collections import OrderedDict
import types
import logging
import numpy as np
import copy
import math
import torchvision.models
import re
from packaging import version

# necessary for PyTorch >= 1.4! See https://github.com/pytorch/pytorch/issues/33463#issuecomment-606399944
class scope_name_workaround(object):
    def __init__(self):
        self.backup = None

    def __enter__(self):
        def _tracing_name(self_, tracing_state):
            if not tracing_state._traced_module_stack:
                return None
            module = tracing_state._traced_module_stack[-1]
            for name, child in module.named_children():
                if child is self_:
                    return name
            return None

        def _slow_forward(self_, *input, **kwargs):
            tracing_state = torch._C._get_tracing_state()
            if not tracing_state or isinstance(self_.forward, torch._C.ScriptMethod):
                return self_.forward(*input, **kwargs)
            if not hasattr(tracing_state, '_traced_module_stack'):
                tracing_state._traced_module_stack = []
            name = _tracing_name(self_, tracing_state)
            if name:
                tracing_state.push_scope('%s[%s]' % (self_._get_name(), name))
            else:
                tracing_state.push_scope(self_._get_name())
            tracing_state._traced_module_stack.append(self_)
            try:
                result = self_.forward(*input, **kwargs)
            finally:
                tracing_state.pop_scope()
                tracing_state._traced_module_stack.pop()
            return result
        
        self.backup = torch.nn.Module._slow_forward
        setattr(torch.nn.Module, '_slow_forward', _slow_forward)

    def __exit__(self, type, value, tb):
        setattr(torch.nn.Module, '_slow_forward', self.backup)

def onnx_name_2_pytorch_name(name):
    name_parts = re.findall('\[.*?\]', name)
    name_parts = [part[1:-1] for part in name_parts]
    return '.'.join(name_parts)

def _hier_flat_dict_build(module, name):
    for n,m in module.named_children():
        if n == name:
            return m
        elif n == name.split('.')[0]:
            return _hier_flat_dict_build(m, '.'.join(name.split('.')[1:]))
    return module

class DeployNode(object):
    def __init__(self, key="", incoming=None, outgoing=None):
        if incoming is None:
            self.incoming = []
        else:
            self.incoming = incoming
        if outgoing is None:
            self.outgoing = []
        else:
            self.outgoing = outgoing
        self.key = key
        self.input_node = False

    def is_input(self):
        return True if len(self.incoming)==0 else False

    def is_output(self):
        return True if len(self.outgoing)==0 else False

    def _traverse_forward(self, fn=None, recurse_max="inf", reduc_fn=lambda ret,x: ret+x, ret_default=0, **kwargs):
        ret = ret_default
        recurse_max = "inf" if recurse_max == "inf" else recurse_max-1
        for o in self.outgoing:
            ret = reduc_fn(ret, fn(o, **kwargs))
            if recurse_max == "inf" or recurse_max > 0:
                ret = reduc_fn(ret, o._traverse_forward(fn, recurse_max=recurse_max, reduc_fn=reduc_fn, **kwargs))
        return ret

    def _traverse_backward(self, fn=None, recurse_max="inf", reduc_fn=lambda ret,x: ret+x, ret_default=0, **kwargs):
        ret = ret_default
        recurse_max = "inf" if recurse_max == "inf" else recurse_max-1
        for o in self.incoming:
            ret = reduc_fn(ret, fn(o, **kwargs))
            if recurse_max == "inf" or recurse_max > 0:
                ret = reduc_fn(ret, o._traverse_backward(fn, recurse_max=recurse_max, reduc_fn=reduc_fn, ret_default=ret_default, **kwargs))
        return ret

class DeployGraph(object):
    def __init__(self, module, dummy_input):
        if version.parse(torch.__version__) < version.parse('1.4.0'):
            trace, _, _ = torch.jit.get_trace_graph(module, dummy_input, _force_outplace=True, _return_inputs_states=True)
            torch.onnx._optimize_trace(trace, torch.onnx.OperatorExportTypes.ONNX)
            graph = trace.graph()
        else:
            with scope_name_workaround():
                graph, _params_dict, _torch_out = torch.onnx.utils._model_to_graph(module, dummy_input, propagate=True, _retain_param_name=True)
        input_dict = {}
        output_dict = {}
        self.non_unique_names_dict = {}
        self.module = module

        # build list of inputs/outputs
        for i,node in enumerate(graph.nodes()):
            op_name = node.scopeName()
            module_name = onnx_name_2_pytorch_name(op_name) + "/" + node.kind().lstrip('::onnx') + "_" + str(i)
            self.non_unique_names_dict[module_name] = onnx_name_2_pytorch_name(op_name)
            for i in node.inputs():
                try:
                    input_dict[i.debugName()].append(module_name)
                except KeyError:
                    input_dict[i.debugName()] = [module_name,]
            for o in node.outputs():
                output_dict[o.debugName()] = module_name

        # build a flat dictionary of modules -- this is centralized in the DeployGraph to ease nemo.transform
        all_nodes = list(set([i for i in output_dict.values()]) | set([i for l in input_dict.values() for i in l]))
        self.module_nodes = OrderedDict([])
        for n in all_nodes:
            nn = n.split("/")[0]
            self.module_nodes[n] = _hier_flat_dict_build(module, nn)
        
        # build a flat dictionary of DeployNodes
        self.nodes = OrderedDict([])
        for n in all_nodes:
            self.nodes[n] = DeployNode(key=n)
            
        # populate outgoing connections
        for ok in output_dict.keys():
            out = output_dict[ok]
            try:
                ils = input_dict[ok]
            except KeyError:
                # unused outputs
                ils = None
            if ils is not None:
                dnl = [self.nodes[i] for i in ils]
                self.nodes[out].outgoing.extend(dnl)

        # populate incoming connections
        nodes_copy = copy.copy(self.nodes) # shallow copy
        for k,n in nodes_copy.items():
            for i,m in enumerate(n.outgoing):
                self.nodes[k].outgoing[i].incoming.append(n)
                
        # identify input nodes (only 1!) FIXME
        self.input_nodes = []
        for k,n in self.nodes.items():
            if len(n.incoming) == 0:
                n.input_node = True
                self.input_nodes.append(n)

        self.jit_graph = graph

    def rebuild_module_dict(self):
        # build a flat dictionary of modules -- this is centralized in the DeployGraph to ease nemo.transform
        for n in self.module_nodes.keys():
            nn = n.split("/")[0]
            self.module_nodes[n] = _hier_flat_dict_build(self.module, nn)

    def print_modules(self):
        for k,n in self.nodes.items():
            print (k, self.module_nodes[k])

    def print_forward_edges(self):
        for k,n in self.nodes.items():
            print (k, [m.key for m in n.outgoing])

    def print_backward_edges(self):
        for k,n in self.nodes.items():
            print (k, [m.key for m in n.incoming])

    def print_jit_graph(self):
        print(self.jit_graph)

    def get_eps_at(self, key, eps_in, use_non_unique_name=True, verbose=False):
        # back-track route to input
        # the procedure is repeated for each incoming edge to the target node
        target = None
        if use_non_unique_name:
            for kk,el in list(self.non_unique_names_dict.items()):
                if el == key:
                    target = self.nodes[kk]
                    break
        else:
            target = self.nodes.get(key, None)
        if target is None:
            print("[nemo-graph] Warning: %s is not a module name" % key)
            # import IPython
            # IPython.embed()
            return None
        eps_list = []
        for incoming_idx in range(len(target.incoming)):
            curr = target
            route = []
            while not curr.is_input():
                if verbose:
                    print("[nemo-graph] backward %s" % (curr.key))
                k = curr.key
                # if current node is the target, use the incoming_idx route
                if curr == target:
                    curr = curr.incoming[incoming_idx]
                else:
                    curr = curr.incoming[0]
                route.append(curr.outgoing.index(self.nodes[k]))
            # forward-track route to node, computing eps
            route = route[::-1]
            eps = eps_in
            for idx in route:
                if verbose:
                    print("[nemo-graph] forward %s %d" % (curr.key, idx))
                if hasattr(self.module_nodes[curr.key], 'get_output_eps'):
                    eps = self.module_nodes[curr.key].get_output_eps(eps)
                try:
                    curr = curr.outgoing[idx]
                except IndexError:
                    print("[nemo-graph] Warning: %s has no outgoing edge" % (curr.key))
                    break
            if type(eps) is float:
                eps_list.append(torch.tensor(eps))
            eps_list.append(eps)
        if len(eps_list) == 1:
            return eps_list[0]
        else:
            return eps_list

    def get_supernodes(self, verbose=False):
        # collect all activation nodes
        actnodes = []
        for k,n in self.nodes.items():
            if isinstance(self.module_nodes[n.key], PACT_Act) or \
               isinstance(self.module_nodes[n.key], PACT_ThresholdAct) or \
               isinstance(self.module_nodes[n.key], PACT_IntegerAct):
                actnodes.append(n)
        supernodes = OrderedDict([])
        # for each activation node, backtrack until another activation node is found
        for target in actnodes:
            # here we assume all activation nodes have only one incoming path, which should be reasonable
            curr = target.incoming[0]
            route = []
            while not isinstance(self.module_nodes[curr.key], PACT_Act) or \
                      isinstance(self.module_nodes[curr.key], PACT_ThresholdAct) or \
                      isinstance(self.module_nodes[curr.key], PACT_IntegerAct):
                route.append((self.non_unique_names_dict[curr.key], self.module_nodes[curr.key]))
                if verbose:
                    print("[nemo-graph] backward %s" % (curr.key))
                try:
                    curr = curr.incoming[0]
                except IndexError:
                    break
            # forward-track route to node
            supernodes[self.non_unique_names_dict[target.key]] = {'supernode': route[::-1], 'previous': self.non_unique_names_dict[curr.key]}
        return supernodes



#
# precision.py
# Francesco Conti <fconti@iis.ee.ethz.ch>
# Alfio Di Mauro <adimauro@iis.ee.ethz.ch>
#
# Copyright (C) 2018-2020 ETH Zurich
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

MIN_REPR_PRECISION = 1e-15
MAX_REPR_SCALE     = 1e+15
MAX_NB_BITS        = 32

class Precision():
    def __init__(self, bits=None, scale=None, positive=False):
        super(Precision, self).__init__()
        self.bits  = bits
        self.scale = scale
        self.positive = positive

    def __gt__(self, other):
        try:
            if self.bits > other.bits:
                return True
            else:
                return False
        except AttributeError:
            if self.bits > other:
                return True
            else:
                return False

    def __lt__(self, other):
        try:
            if self.bits < other.bits:
                return True
            else:
                return False
        except AttributeError:
            if self.bits < other:
                return True
            else:
                return False

    def __ge__(self, other):
        try:
            if self.bits >= other.bits:
                return True
            else:
                return False
        except AttributeError:
            if self.bits >= other:
                return True
            else:
                return False

    def __le__(self, other):
        try:
            if self.bits <= other.bits:
                return True
            else:
                return False
        except AttributeError:
            if self.bits <= other:
                return True
            else:
                return False

    def __eq__(self, other):
        try:
            if self.bits == other.bits:
                return True
            else:
                return False
        except AttributeError:
            if self.bits == other:
                return True
            else:
                return False

    def __ne__(self, other):
        try:
            if self.bits != other.bits:
                return True
            else:
                return False
        except AttributeError:
            if self.bits != other:
                return True
            else:
                return False

    def set_eps(self, eps):
        self.bits = np.log2(-eps)

    def set_bits(self, bits):
        self.bits = bits

    def set_clip(self, clip):
        self.scale = clip

    def set_scale(self, scale):
        self.scale = scale
    
    def get_eps(self):
        if self.bits is None or self.scale is None:
            return MIN_REPR_PRECISION
        if not self.positive:
            return 2.0**(-(self.bits-1)) * self.scale
        else:
            return 2.0**(-self.bits) * self.scale

    def get_clip(self):
        return self.get_scale()

    def get_scale(self):
        if self.scale is None:
            return MAX_REPR_SCALE
        return self.scale

    def get_bits(self):
        if self.bits is None:
            return MAX_NB_BITS
        return self.bits



#
# evaluation.py
# Francesco Conti <fconti@iis.ee.ethz.ch>
# Alfio Di Mauro <adimauro@iis.ee.ethz.ch>
#
# Copyright (C) 2018-2020 ETH Zurich
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import logging
import nemo
import torch
from collections import OrderedDict
from tqdm import tqdm

class EvaluationEngine():
    def __init__(self, net, validate_fn=None, validate_data=None, precision_rule=None, min_prec_dict=None):
        super(EvaluationEngine, self).__init__()

        self.precision_rule = precision_rule
        if self.precision_rule is None:
            self.scale_activations = True
            self.scale_weights = True
        self.min_prec_dict = min_prec_dict
        self.loss_prev = 1e3
        self.net = net
        self.validate_fn = validate_fn
        self.validate_data = validate_data
        self.reset_grids()

    def reset_grids(self, W_start=None, W_stop=None, W_step=None, x_start=None, x_stop=None, x_step=None):
        first_W = self.precision_rule['0']['W_bits'] if W_start is None else W_start
        first_x = self.precision_rule['0']['x_bits'] if x_start is None else x_start
        step_W = float(self.precision_rule['W_bit_scaler']) if W_step is None else W_step
        step_x = float(self.precision_rule['x_bit_scaler']) if x_step is None else x_step
        last_W = float(self.precision_rule['W_bit_stop_condition'])+step_W if W_stop is None else W_stop
        last_x = float(self.precision_rule['x_bit_stop_condition'])+step_x if x_stop is None else x_stop
        W = np.arange(first_W, last_W, step_W)
        x = np.arange(first_x, last_x, step_x)
        if len(W) == 0:
            W = np.asarray([first_W,])
        if len(x) == 0:
            x = np.asarray([first_x,])
        logging.info("[Evaluation]\t Setting up new grid: W=%s x=%s" % (W, x))
        wgrid, xgrid = np.meshgrid(W, x)
        self.acc = np.zeros_like(wgrid, dtype='float32')
        self.wgrid = wgrid.flatten()
        self.xgrid = xgrid.flatten()
        self.idx = -1

    def report(self, acc):
        sh = self.acc.shape
        self.acc = self.acc.flatten()
        self.acc[self.idx] = acc.item()
        self.acc = self.acc.reshape(sh)

    def __suffix(self):
        return "_%.1fx%.1fb" % (self.net.W_precision.get_bits(), self.net.x_precision.get_bits())

    def step(self, checkpoint_name='checkpoint', verbose=False):
        if self.precision_rule is None:
            return
        try:
            curr_regime = self.precision_rule[0]
        except KeyError:
            try:
               curr_regime = self.precision_rule[str(0)]
            except KeyError:
               curr_regime = None

        if self.idx == self.wgrid.shape[0] - 1:
            return False

        self.idx += 1
        self.net.change_precision(bits=self.wgrid[self.idx], scale_activations=False, scale_weights=True, reset_alpha=False, verbose=verbose, min_prec_dict=self.min_prec_dict)
        self.net.change_precision(bits=self.xgrid[self.idx], scale_activations=True, scale_weights=False, reset_alpha=False, verbose=verbose, min_prec_dict=self.min_prec_dict)
        return True

    # the rationale here is to define frontiers using a certain threshold, e.g. 90% of the top accuracy
    # reached by pure evaluation. configurations within the frontier are considered 'good enough'
    # and not worth being considered for retraining (only minor fine-tuning is necessary).
    def get_next_config(self, upper_threshold=0.9, strategy='min_precision', verbose=False, Wbits_curr=None, xbits_curr=None, timeout=25):
        def create_bins(upper_threshold):
            bins = np.asarray([0, 1.0-upper_threshold, upper_threshold]) * self.acc.max()
            acc_idx = np.digitize(self.acc, bins)
            return acc_idx
        acc_idx = create_bins(upper_threshold)
        for i in range(timeout):
            if len(acc_idx[acc_idx == 2]) == 0:
                # exponentially increase threshold
                upper_threshold /= upper_threshold
                if verbose:
                    logging.info("[Evaluation]\t No middle-bin element, changing threshold to %.3e." % upper_threshold)
                acc_idx = create_bins(upper_threshold)
        if len(acc_idx[acc_idx == 2]) == 0:
            return self.wgrid.reshape(self.acc.shape)[0,0], self.xgrid.reshape(self.acc.shape)[0,0]
        if verbose:
            logging.info("[Evaluation]\t Top-bin:    %s" % (np.dstack((self.wgrid.reshape(self.acc.shape)[acc_idx == 3], self.xgrid.reshape(self.acc.shape)[acc_idx == 3], self.acc[acc_idx == 3]))))
            logging.info("[Evaluation]\t Middle-bin: %s" % (np.dstack((self.wgrid.reshape(self.acc.shape)[acc_idx == 2], self.xgrid.reshape(self.acc.shape)[acc_idx == 2], self.acc[acc_idx == 2]))))
            logging.info("[Evaluation]\t Bottom-bin: %s" % (np.dstack((self.wgrid.reshape(self.acc.shape)[acc_idx == 1], self.xgrid.reshape(self.acc.shape)[acc_idx == 1], self.acc[acc_idx == 1]))))
        if strategy == 'max_accuracy':
            if verbose:
                logging.info("[Evaluation]\t Select the most accurate one from the middle-bin.")
            idxs = np.unravel_index(self.acc[acc_idx == 2].argmax(), self.acc.shape)
        elif strategy == 'min_precision': 
            if verbose:
                logging.info("[Evaluation]\t Select the lowest precision one from the middle-bin.")
            xmin = self.xgrid.reshape(self.acc.shape)[acc_idx == 2].min()
            mask = np.logical_and(acc_idx == 2, self.xgrid.reshape(self.acc.shape)==xmin)
            wtmpgrid = np.copy(self.wgrid.reshape(self.acc.shape))
            wtmpgrid[np.logical_not(mask)] = 1e6
            idxs = np.unravel_index(wtmpgrid.argmin(), self.acc.shape)
        Widx = idxs[1]
        xidx = idxs[0]
        if verbose:
            logging.info("[Evaluation]\t Returning idxs: %d,%d prec %d,%d" % (Widx, xidx, self.wgrid[Widx], self.xgrid[xidx]))
        return self.wgrid.reshape(self.acc.shape)[0,Widx], self.xgrid.reshape(self.acc.shape)[xidx,0]



#
# pact.py
# Francesco Conti <fconti@iis.ee.ethz.ch>
# Alfio Di Mauro <adimauro@iis.ee.ethz.ch>
#
# Copyright (C) 2018-2020 ETH Zurich
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding=utf-8
import math
import torch
from torch.nn.modules.utils import _single,_pair
from nemo.precision import Precision
import numpy as np

import logging

# Create custom symbolic function
from torch.onnx.symbolic_helper import parse_args

DEFAULT_ACT_REQNT_FACTOR  = 32
DEFAULT_ADD_REQNT_FACTOR  = 32
DEFAULT_POOL_REQNT_FACTOR = 32
DEFAULT_QBATCHNORM_PREC = 12
QD_REQUANT_DEBUG = False

__all__ = ["PACT_Conv1d", "PACT_Conv2d", "PACT_Linear", "PACT_Act", "PACT_ThresholdAct", "PACT_IntegerAct", "PACT_IntegerAvgPool2d", "PACT_Identity", "PACT_QuantizedBatchNormNd", "PACT_IntegerBatchNormNd"]

# re-quantize from a lower precision (larger eps_in) to a higher precision (lower eps_out)
# requantization rounding can be excluded for debug purposes, e.g., to identify numerical
# differences that are hidden by the requantization approximation
def pact_quantized_requantize(t, eps_in, eps_out, D=1, exclude_requant_rounding=QD_REQUANT_DEBUG):
    if exclude_requant_rounding:
        return torch.floor(t / eps_out)
    else:
        eps_ratio = (D*eps_in/eps_out).round()
        return torch.floor(t / eps_in * eps_ratio / D)

# re-quantize from a lower precision (larger eps_in) to a higher precision (lower eps_out)
def pact_integer_requantize(t, eps_in, eps_out, D=1):
    eps_ratio = (D*eps_in/eps_out).round()
    device = t.device
    return torch.as_tensor((torch.as_tensor(t, dtype=torch.int64) * torch.as_tensor(eps_ratio, dtype=torch.int64) / torch.as_tensor(D, dtype=torch.int64)), dtype=torch.float32, device=device)

# re-quantize from a lower precision (larger eps_in) to a higher precision (lower eps_out)
def pact_integer_requantize_add(*t, eps_in_list, eps_out, D=1):
    # unrolling the first iteration of the loop instead of using torch.zeros to init y is necessary for correct ONNX export
    eps_in = eps_in_list[0]
    eps_ratio = (D*eps_in/eps_out).round()
    y = t[0] * eps_ratio
    for i in range(1,len(eps_in_list)):
        eps_in = eps_in_list[i]
        eps_ratio = (D*eps_in/eps_out).round()
        y += t[i] * eps_ratio
    y = (y / D).floor()
    return y

# PACT quantization for inference
def pact_quantize_inference(W, eps, clip):
    W_quant = W.clone().detach()
    W_quant.data[:] = (W_quant.data[:] / eps).floor()*eps
    W_quant.clamp_(0, clip.item())
    return W_quant

def pact_quantize_deploy(W, eps, clip):
    W = (W / eps).floor()*eps
    W = W.clamp(0, clip)
    return W

# PACT signed quantization for inference (workaround for pact_quantize_signed not functional in inference)
def pact_quantize_signed_inference(W, eps, clip):
    return pact_quantize_asymm_inference(W, torch.as_tensor(eps), torch.as_tensor(clip), torch.as_tensor(clip))

# PACT asymmetric quantization for inference (workaround for pact_quantize_asymm not functional in inference)
def pact_quantize_asymm_inference(W, eps, alpha, beta, train_loop=True, train_loop_oldprec=None):
    # for numerical reasons, W_quant should be put at a "small_eps" of its "pure" value to enable
    # running againt the weights through pact_quantize_asymm_inference (the torch.floor function
    # won't return the correct value otherwise)
    # we choose small_eps = eps/2
    if not train_loop and train_loop_oldprec is not None:
        W_quant = W.clone().detach()
        W_quant.data[:] = (W_quant.data[:] / train_loop_oldprec).floor()*train_loop_oldprec + eps*0.5
    else:
        W_quant = W.clone().detach() + eps*0.5
    W_quant.data[:] = (W_quant.data[:] / eps).floor()*eps
    # alpha, beta are also represented with quantized numbers
    alpha = torch.ceil(alpha/eps)*eps
    beta  = torch.floor(beta/eps)*eps
    W_quant.clamp_(-alpha.item(), beta.item() + eps.item())
    return W_quant

# DEPRECATED
def pact_pwl(x, eps, alpha, beta, q0=0):
    beta = beta.abs()
    beta_cumsum = beta.cumsum(0)
    m = eps / beta # at initial PWL step, should be 1!
    # compute step, extended with -alpha at the beginning
    step = torch.cat((torch.as_tensor((-np.infty,), device=alpha.device), -alpha, -alpha + beta_cumsum[:-1], torch.as_tensor((+np.infty,), device=alpha.device)))
    sr = torch.arange(len(step.shape)+len(x.shape))
    sr_m1  = sr[-1].clone().detach()
    sr[1:] = sr[:-1].clone().detach()
    sr[0]  = sr_m1
    step = step.repeat((*x.shape,1)).permute(tuple(sr.numpy()))
    # compute q offset
    q = torch.zeros_like(beta)
    q[1:] = q0 + eps * (torch.arange(1, beta.shape[0], dtype=torch.float32, device=alpha.device) - 1 - (-alpha + beta_cumsum[:-1])/beta[1:])
    # extend m with a final 0
    m = torch.cat((torch.ones(1, device=alpha.device), m))
    q = torch.cat((torch.zeros(1, device=alpha.device), q))
    y = torch.zeros_like(x)
    # compare
    inside = ((x >= step[:-1]) * (x < step[1:]))
    for i in range(len(inside.shape)-1):
        m = torch.unsqueeze(m, -1)
        q = torch.unsqueeze(q, -1)
    x = torch.unsqueeze(x, 0)
    y = torch.where(inside, m*x+q, y).sum(0)
    del inside, m, q, beta_cumsum
    return y.clamp(-alpha.item(), alpha.item()-eps.item())

# PACT activation: https://arxiv.org/pdf/1805.06085.pdf
class PACT_QuantFunc(torch.autograd.Function):
    r"""PACT (PArametrized Clipping acTivation) quantization function for activations.

        Implements a :py:class:`torch.autograd.Function` for quantizing activations in :math:`Q` bits using the PACT strategy.
        In forward propagation, the function is defined as 
        
        .. math::
            \mathbf{y} = f(\mathbf{x}) = 1/\varepsilon \cdot \left\lfloor\mathrm{clip}_{ [0,\alpha) } (\mathbf{x})\right\rfloor \cdot \varepsilon
        
        where :math:`\varepsilon` is the quantization precision:
        
        .. math::
            \varepsilon = \alpha / (2^Q - 1)
        
        In backward propagation, using the Straight-Through Estimator, the gradient of the function is defined as
        
        .. math::
            \mathbf{\nabla}_\mathbf{x} \mathcal{L} &\doteq \mathbf{\nabla}_\mathbf{y} \mathcal{L}
        
        It can be applied by using its static `.apply` method:
    
    :param input: the tensor containing :math:`x`, the activations to be quantized.
    :type  input: `torch.Tensor`
    :param eps: the precomputed value of :math:`\varepsilon`.
    :type  eps: `torch.Tensor` or float
    :param alpha: the value of :math:`\alpha`.
    :type  alpha: `torch.Tensor` or float
    :param delta: constant to sum to `eps` for numerical stability (default unused, 0 ).
    :type  delta: `torch.Tensor` or float
    
    :return: The quantized input activations tensor.
    :rtype:  `torch.Tensor`

    """

    @staticmethod
    def forward(ctx, input, eps, alpha, delta=0):
        where_input_nonclipped = (input >= 0) * (input < alpha)
        where_input_gtalpha = (input >= alpha)
        ctx.save_for_backward(where_input_nonclipped, where_input_gtalpha)
        return ((input / (eps+delta)).floor() * eps).clamp(0., alpha.data[0]-eps.data[0])

    @staticmethod
    def backward(ctx, grad_output):
        # see Hubara et al., Section 2.3
        where_input_nonclipped, where_input_gtalpha = ctx.saved_variables
        zero = torch.zeros(1).to(where_input_nonclipped.device)
        grad_input = torch.where(where_input_nonclipped, grad_output, zero)
        grad_alpha = torch.where(where_input_gtalpha, grad_output, zero).sum().expand(1)
        return grad_input, None, grad_alpha

pact_quantize = PACT_QuantFunc.apply

# DEPRECATED
class PACT_QuantFunc_Signed(torch.autograd.Function):
    r"""PACT (PArametrized Clipping acTivation) quantization function for weights (simmetric).

        Implements a :py:class:`torch.autograd.Function` for quantizing weights in :math:`Q` bits using a symmetric PACT-like strategy (original
        PACT is applied only to activations, using DoReFa-style weights).
        In forward propagation, the function is defined as 
        
        .. math::
            \mathbf{y} = f(\mathbf{x}) = 1/\varepsilon \cdot \left\lfloor\mathrm{clip}_{ [-\alpha,+\alpha) } (\mathbf{x})\right\rfloor \cdot \varepsilon
        
        where :math:`\varepsilon` is the quantization precision:
        
        .. math::
            \varepsilon = 2\cdot\alpha / (2^Q - 1)
        
        In backward propagation, using the Straight-Through Estimator, the gradient of the function is defined as
        
        .. math::
            \mathbf{\nabla}_\mathbf{x} \mathcal{L} &\doteq \mathbf{\nabla}_\mathbf{y} \mathcal{L}
        
        It can be applied by using its static `.apply` method:
    
    :param input: the tensor containing :math:`x`, the weights to be quantized.
    :type  input: `torch.Tensor`
    :param eps: the precomputed value of :math:`\varepsilon`.
    :type  eps: `torch.Tensor` or float
    :param alpha: the value of :math:`\alpha`.
    :type  alpha: `torch.Tensor` or float
    :param delta: constant to sum to `eps` for numerical stability (default 0, unused).
    :type  delta: `torch.Tensor` or float
    
    :return: The quantized weights tensor.
    :rtype:  `torch.Tensor`

    """

    @staticmethod
    def forward(ctx, input, eps, alpha, delta=0):
        where_input_nonclipped = (input >= -alpha) * (input < alpha)
        where_input_gtalpha = (input >= alpha) + (input < -alpha)
        ctx.save_for_backward(where_input_nonclipped, where_input_gtalpha)
        return (input.clamp(-alpha.data[0], alpha.data[0]) / (eps+delta)).floor() * eps

    @staticmethod
    def backward(ctx, grad_output):
        # see Hubara et al., Section 2.3
        where_input_nonclipped, where_input_gtalpha = ctx.saved_variables
        zero = torch.zeros(1).to(where_input_nonclipped.device)
        grad_input = torch.where(where_input_nonclipped, grad_output, zero)
        grad_alpha = torch.where(where_input_gtalpha, grad_output, zero).sum().expand(1)
        return grad_input, None, grad_alpha

# DEPRECATED
pact_quantize_signed = PACT_QuantFunc_Signed.apply

class PACT_QuantFunc_Asymm(torch.autograd.Function):
    r"""PACT (PArametrized Clipping acTivation) quantization function (asymmetric).

        Implements a :py:class:`torch.autograd.Function` for quantizing weights in :math:`Q` bits using an asymmetric PACT-like strategy (original
        PACT is applied only to activations, using DoReFa-style weights).
        In forward propagation, the function is defined as 
        
        .. math::
            \mathbf{y} = f(\mathbf{x}) = 1/\varepsilon \cdot \left\lfloor\mathrm{clip}_{ [-\alpha,+\beta) } (\mathbf{x})\right\rfloor \cdot \varepsilon
        
        where :math:`\varepsilon` is the quantization precision:
        
        .. math::
            \varepsilon = (\alpha+\beta) / (2^Q - 1)
        
        In backward propagation, using the Straight-Through Estimator, the gradient of the function is defined as
        
        .. math::
            \mathbf{\nabla}_\mathbf{x} \mathcal{L} &\doteq \mathbf{\nabla}_\mathbf{y} \mathcal{L}
        
        It can be applied by using its static `.apply` method:
    
    :param input: the tensor containing :math:`x`, the weights to be quantized.
    :type  input: `torch.Tensor`
    :param eps: the precomputed value of :math:`\varepsilon`.
    :type  eps: `torch.Tensor` or float
    :param alpha: the value of :math:`\alpha`.
    :type  alpha: `torch.Tensor` or float
    :param beta: the value of :math:`\beta`.
    :type  beta: `torch.Tensor` or float
    :param delta: constant to sum to `eps` for numerical stability (default unused, 0).
    :type  delta: `torch.Tensor` or float
    
    :return: The quantized weights tensor.
    :rtype:  `torch.Tensor`

    """

    @staticmethod
    def forward(ctx, input, eps, alpha, beta, delta=0):
        # we quantize also alpha, beta. for beta it's "cosmetic", for alpha it is 
        # substantial, because also alpha will be represented as a wholly integer number
        # down the line
        alpha_quant = (alpha.item() / (eps+delta)).ceil()  * eps
        beta_quant  = (beta.item()  / (eps+delta)).floor() * eps
        where_input_nonclipped = (input >= -alpha_quant) * (input < beta_quant)
        where_input_ltalpha = (input < -alpha_quant)
        where_input_gtbeta = (input >= beta_quant)
        ctx.save_for_backward(where_input_nonclipped, where_input_ltalpha, where_input_gtbeta)
        return (input.clamp(-alpha_quant.item(), beta_quant.item()) / (eps+delta)).round() * eps

    @staticmethod
    def backward(ctx, grad_output):
        # see Hubara et al., Section 2.3
        where_input_nonclipped, where_input_ltalpha, where_input_gtbeta = ctx.saved_variables
        zero = torch.zeros(1).to(where_input_nonclipped.device)
        grad_input = torch.where(where_input_nonclipped, grad_output, zero)
        grad_alpha = torch.where(where_input_ltalpha, grad_output, zero).sum().expand(1)
        grad_beta  = torch.where(where_input_gtbeta, grad_output, zero).sum().expand(1)
        return grad_input, None, grad_alpha, grad_beta

pact_quantize_asymm = PACT_QuantFunc_Asymm.apply

class PACT_Act(torch.nn.Module):
    r"""PACT (PArametrized Clipping acTivation) activation.

    Implements a :py:class:`torch.nn.Module` to implement PACT-style activations. It is meant to replace :py:class:`torch.nn.ReLU`, :py:class:`torch.nn.ReLU6` and
    similar activations in a PACT-quantized network.

    This layer can also operate in a special mode, defined by the `statistics_only` member, in which the layer runs in
    forward-prop without quantization, collecting statistics on the activations that can then be
    used to reset the value of :math:`\alpha`.
    In this mode, the layer collects:
    - tensor-wise maximum value ever seen
    - running average with momentum 0.9
    - running variance with momentum 0.9

    """

    def __init__(self, precision=None, alpha=1., backprop_alpha=True, statistics_only=False, leaky=None, requantization_factor=DEFAULT_ACT_REQNT_FACTOR):
        r"""Constructor. Initializes a :py:class:`torch.nn.Parameter` for :math:`\alpha` and sets
            up the initial value of the `statistics_only` member.

        :param precision: instance defining the current quantization level (default `None`).
        :type  precision: :py:class:`nemo.precision.Precision`
        :param alpha: the value of :math:`\alpha`.
        :type  alpha: `torch.Tensor` or float
        :param backprop_alpha: default `True`; if `False`, do not update the value of `\alpha` with backpropagation.
        :type  backprop_alpha: bool
        :param statistics_only: initialization value of `statistics_only` member.
        :type  statistics_only: bool

        """

        super(PACT_Act, self).__init__()
        if precision is None:
            self.precision = Precision()
        else:
            self.precision = Precision(bits=precision.get_bits())
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.alpha = torch.nn.Parameter(torch.Tensor((alpha,)).to(device), requires_grad=backprop_alpha)
        self.alpha_p = alpha
        self.statistics_only = statistics_only
        self.deployment = False
        self.eps_in = None
        self.leaky = leaky
        self.requantization_factor = requantization_factor

        # these are only used to gather statistics
        self.max          = torch.nn.Parameter(torch.zeros_like(self.alpha.data).to(device), requires_grad=False)
        self.min          = torch.nn.Parameter(torch.zeros_like(self.alpha.data).to(device), requires_grad=False)
        self.running_mean = torch.nn.Parameter(torch.zeros_like(self.alpha.data).to(device), requires_grad=False)
        self.running_var  = torch.nn.Parameter(torch.ones_like(self.alpha.data).to(device),  requires_grad=False)

        self.precise = False

    def set_static_precision(self, limit_at_32_bits=True, **kwargs):
        r"""Sets static parameters used only for deployment.

        """
        # item() --> conversion to float
        # apparently causes a slight, but not invisibile, numerical divergence
        # between FQ and QD stages
        self.eps_static   = self.alpha.clone().detach()/(2.0**(self.precision.get_bits())-1)
        self.alpha_static = self.alpha.clone().detach()
        # D is selected as a power-of-two
        D = 2.0**torch.ceil(torch.log2(self.requantization_factor * self.eps_static / self.eps_in))
        if not limit_at_32_bits:
            self.D = D
        else:
            self.D = min(D, 2.0**(32-1-(self.precision.get_bits())))

    def get_output_eps(self, eps_in):
        r"""Get the output quantum (:math:`\varepsilon`) given the input one.

        :param eps_in: input quantum :math:`\varepsilon_{in}`.
        :type  eps_in: :py:class:`torch.Tensor`
        :return: output quantum :math:`\varepsilon_{out}`.
        :rtype:  :py:class:`torch.Tensor`

        """

        return self.alpha/(2.0**(self.precision.get_bits())-1)

    def reset_alpha(self, use_max=True, nb_std=5.):
        r"""Reset the value of :math:`\alpha`. If `use_max` is `True`, then the highest tensor-wise value collected
            in the statistics collection phase is used. If `False`, the collected standard deviation multiplied by
            `nb_std` is used as a parameter

        :param use_max: if True, use the tensor-wise maximum value collected in the statistics run as new :math:`\alpha` (default True).
        :type  use_max: bool
        :param nb_std: number of standard deviations to be used to initialize :math:`\alpha` if `use_max` is False.
        :type  nb_std: float

        """

        if use_max:
            self.alpha.data[0] = self.max.item()
        else:
            self.alpha.data[0] = nb_std * torch.sqrt(self.running_var).item()

    def get_statistics(self):
        r"""Returns the statistics collected up to now.
    
        :return: The collected statistics (maximum, running average, running variance).
        :rtype:  tuple of floats

        """
        return self.max.item(), self.running_mean.item(), self.running_var.item()
    
    def forward(self, x):
        r"""Forward-prop function for PACT-quantized activations.
        
        See :py:class:`nemo.quant.pact_quant.PACT_QuantFunc` for details on the normal operation performed by this layer.
        In statistics mode, it uses a normal ReLU and collects statistics in the background.

        :param x: input activations tensor.
        :type  x: :py:class:`torch.Tensor`
        
        :return: output activations tensor.
        :rtype:  :py:class:`torch.Tensor`

        """

        if self.deployment:
            x_rq = pact_quantized_requantize(x, self.eps_in, self.eps_static, self.D, exclude_requant_rounding=self.precise) * self.eps_static
            return x_rq.clamp(0, self.alpha_static.data[0])
        elif self.statistics_only:
            if self.leaky is None:
                x = torch.nn.functional.relu(x)
            else:
                x = torch.nn.functional.leaky_relu(x, self.leaky)
            with torch.no_grad():
                self.max[:] = max(self.max.item(), x.max())
                self.min[:] = min(self.min.item(), x.min())
                self.running_mean[:] = 0.9 * self.running_mean.item() + 0.1 * x.mean()
                self.running_var[:]  = 0.9 * self.running_var.item()  + 0.1 * x.std()*x.std()
            return x
        else:
            eps = self.alpha/(2.0**(self.precision.get_bits())-1)
            return pact_quantize(x, eps, self.alpha + eps)

class PACT_IntegerAdd(torch.nn.Module):
    r"""PACT (PArametrized Clipping acTivation) activation for integer images.

    Implements a :py:class:`torch.nn.Module` to implement PACT-style activations. It is meant to replace :py:class:`nemo.quant.pact_quant.PACT_Act`
    in a integer network. It is not meant to be trained but rather generated by a NeMO transformation. 

    """

    def __init__(self, alpha=1., precision=None, requantization_factor=DEFAULT_ADD_REQNT_FACTOR):
        r"""Constructor. Initializes a :py:class:`torch.nn.Parameter` for :math:`\alpha`.

        :param precision: instance defining the current quantization level (default `None`).
        :type  precision: :py:class:`nemo.precision.Precision`
        :param alpha: the value of :math:`\alpha`.
        :type  alpha: `torch.Tensor` or float
        :param requantization_factor: minimum target ratio in requantization fraction (default 32).
        :type  requantization_factor: int

        """
        super(PACT_IntegerAdd, self).__init__()
        if precision is None:
            self.precision = Precision(bits=8)
        else:
            self.precision = Precision(bits=precision.get_bits())
        device = 'cuda' if torch.cuda.is_available() else 'cpu'

        self.requantization_factor = requantization_factor
        self.deployment = False
        self.integerized = False
        self.eps_in_list = []
    
    def get_output_eps(self, eps_in_list):
        if type(eps_in_list) is list:
            self.eps_in_list = eps_in_list
        # for now, select the biggest eps_out as the target one
        self.eps_out = max(self.eps_in_list)
        self.alpha_out = 2.0**(self.precision.get_bits())-1
        # D is selected as a power-of-two
        self.D = 2.0**torch.ceil(torch.log2(self.requantization_factor * self.eps_out / min(self.eps_in_list)))
        return self.eps_out

    def forward(self, *x):
        if not self.deployment or not self.integerized:
            y = x[0]
            for i in range(1,len(x)):
                y += x[i]
            return y
        else:
            return pact_integer_requantize_add(*x, eps_in_list=self.eps_in_list, eps_out=self.eps_out, D=self.D)

class PACT_IntegerAvgPool2d(torch.nn.AvgPool2d):
    def __init__(self, kernel_size, stride=None, padding=0, ceil_mode=False,
            count_include_pad=True, divisor_override=None, requantization_factor=DEFAULT_POOL_REQNT_FACTOR):
        super(PACT_IntegerAvgPool2d, self).__init__(kernel_size, stride=stride, padding=padding, ceil_mode=ceil_mode,
            count_include_pad=count_include_pad, divisor_override=divisor_override)
        self.requantization_factor = requantization_factor
        if type(self.kernel_size) is not tuple and type(self.kernel_size) is not list:
            ks = (self.kernel_size, self.kernel_size)
        else:
            ks = self.kernel_size
        self.kernel_size = ks
        self.D = 2.0**torch.ceil(torch.log2(self.requantization_factor*torch.as_tensor(1.0*self.kernel_size[0]*self.kernel_size[1])))

    def get_output_eps(self, eps_in):
        r"""Get the output quantum (:math:`\varepsilon`) given the input one.

        :param eps_in: input quantum :math:`\varepsilon_{in}`.
        :type  eps_in: :py:class:`torch.Tensor`
        :return: output quantum :math:`\varepsilon_{out}`.
        :rtype:  :py:class:`torch.Tensor`

        """
        return eps_in

    def forward(self, input):
        y = torch.nn.functional.avg_pool2d(input, self.kernel_size, self.stride,
                                           self.padding, self.ceil_mode, self.count_include_pad, self.divisor_override)
        if self.divisor_override is not None:
            y *= self.divisor_override
        else:
            y *= self.kernel_size[0] * self.kernel_size[1]
        ratio = (self.D/(self.kernel_size[0]*self.kernel_size[1])).round()
        return (y * ratio / self.D).floor()

class PACT_IntegerAct(torch.nn.Module):
    r"""PACT (PArametrized Clipping acTivation) activation for integer images.

    Implements a :py:class:`torch.nn.Module` to implement PACT-style activations. It is meant to replace :py:class:`nemo.quant.pact_quant.PACT_Act`
    in a integer network. It is not meant to be trained but rather generated by a NeMO transformation. 

    """

    def __init__(self, eps_in, eps_out, alpha=1., precision=None, requantization_factor=DEFAULT_ACT_REQNT_FACTOR, **kwargs):
        r"""Constructor. Initializes a :py:class:`torch.nn.Parameter` for :math:`\alpha`.

        :param eps_in: input quantum.
        :type  eps_in: `torch.Tensor` or float
        :param eps_out: output quantum.
        :type  eps_out: `torch.Tensor` or float
        :param alpha: the value of :math:`\alpha`.
        :type  alpha: `torch.Tensor` or float
        :param precision: instance defining the current quantization level (default `None`).
        :type  precision: :py:class:`nemo.precision.Precision`
        :param requantization_factor: minimum target ratio in requantization fraction (default 32).
        :type  requantization_factor: int

        """
        super(PACT_IntegerAct, self).__init__()
        if precision is None:
            self.precision = Precision(bits=16)
        else:
            self.precision = Precision(bits=precision.get_bits())
        device = 'cuda' if torch.cuda.is_available() else 'cpu'

        self.alpha = torch.nn.Parameter(alpha)

        self.eps_in = eps_in
        self.eps_out = eps_out
        self.requantization_factor = requantization_factor

    def set_output_eps(self, limit_at_32_bits=True, **kwargs):
        r"""Sets static parameters used in deployment mode.

        :param limit_at_32_bits: If True (default), define divider D so that a 32-bit accumulator can be used.
        :type  limit_at_32_bits: bool

        """
        # self.eps_out   = self.alpha.item()/(2.0**(self.precision.get_bits())-1)
        self.alpha_out = 2.0**(self.precision.get_bits())-1
        # D is selected as a power-of-two
        D = 2.0**torch.ceil(torch.log2(self.requantization_factor * self.eps_out / self.eps_in))
        if not limit_at_32_bits:
            self.D = D
        else:
            self.D = min(D, 2.0**(32-(self.precision.get_bits())))

    def get_output_eps(self, eps_in):
        r"""Get the output quantum (:math:`\varepsilon`) given the input one.

        :param eps_in: input quantum :math:`\varepsilon_{in}`.
        :type  eps_in: :py:class:`torch.Tensor`
        :return: output quantum :math:`\varepsilon_{out}`.
        :rtype:  :py:class:`torch.Tensor`

        """

        return self.eps_out

    def forward(self, x):
        r"""Forward-prop function for integer PACT activations.
        
        In forward propagation, the function is defined as 
        
        .. math::
            \mathbf{y} = \mathrm{clip}_{[0,\alpha/\varepsilon_\mathbf{y}]}\left( \left\lfloor\frac{\varepsilon_\mathbf{x}\cdot 2^d}{\varepsilon_\mathbf{y}}\right\rfloor \cdot {q}_\mathbf{x}(\mathbf{x}) \gg d \right)
        
        It performs a requantization from :math:`\varepsilon_\mathbf{y}` to :math:`\varepsilon_\mathbf{y}`.
        
        This layer is not meant to be trained directly, and as such using it in backward is not recommended.

        :param x: input activations tensor.
        :type  x: :py:class:`torch.Tensor`
        
        :return: output activations tensor.
        :rtype:  :py:class:`torch.Tensor`

        """

        x_rq = pact_integer_requantize(x, self.eps_in, self.eps_out, self.D)
        return x_rq.clamp(0, self.alpha_out)

class PACT_ThresholdAct(torch.nn.Module):
    r"""PACT (PArametrized Clipping acTivation) activation with thresholding.

    Implements a :py:class:`torch.nn.Module` to implement PACT-style activations. It is meant to replace :py:class:`nemo.quant.pact_quant.PACT_Act`
    if BatchNorms are merged with PACT_Act activations. It works both in fake-quantized and integer mode.
    It is not meant to be trained but rather generated by a NeMO transformation.
    Thresholds are not directly memorized; instead, they are split in two parameters :math:`\kappa` and :math:`\lambda`. If :math:`p` is an iterator
    on the number of thresholds,

    .. math::
        \tau_p &= p \cdot \sigma/\gamma\varepsilon - \sigma/\gamma*\beta + \mu \\
        \tau_p &= p \cdot \kappa\varepsilon + \lambda \\
        \kappa &= \sigma/\gamma, \lambda = \mu - \sigma/\gamma*\beta
    
    After integerization:

    .. math::
        \tau_p = 1/\varepsilon_x/\varepsilon_W \cdot (p \cdot \kappa\varepsilon + \lambda)

    """

    def __init__(self, precision=None, alpha=1., nb_channels=1):
        r"""Constructor. Initializes a :py:class:`torch.nn.Parameter` for :math:`\alpha` and the target
        number of channels.

        :param precision: instance defining the current quantization level (default `None`).
        :type  precision: :py:class:`nemo.precision.Precision`
        :param alpha: the value of :math:`\alpha`.
        :type  alpha: `torch.Tensor` or float

        """
        super(PACT_ThresholdAct, self).__init__()
        if precision is None:
            self.precision = Precision()
        else:
            self.precision = Precision(bits=precision.get_bits())
        device = 'cuda' if torch.cuda.is_available() else 'cpu'

        self.alpha = torch.nn.Parameter(alpha)

        # thresholds are memorized in the format:
        #   tau_hat(p) = sigma/gamma*eps * p - sigma/gamma*beta + mu
        #   tau_hat(p) = kappa*eps * p + lamda
        #   kappa = sigma/gamma, lamda = mu - sigma/gamma*beta
        # with p in [0,1,...,2**precision.bits()-1]
        # INT-Q thresholds are:
        #   tau(p) = 1/eps_x/eps_W * (kappa * eps * p + lambda)
        self.kappa = torch.nn.Parameter(torch.zeros(nb_channels).to(device))
        self.lamda = torch.nn.Parameter(torch.zeros(nb_channels).to(device))

    def forward(self, x):
        r"""Forward-prop function for threshold-based PACT activations.
        
        This layer is not meant to be trained directly, and as such using it in backward is not recommended.

        :param x: input activations tensor.
        :type  x: :py:class:`torch.Tensor`
        
        :return: output activations tensor.
        :rtype:  :py:class:`torch.Tensor`

        """
        kappa = self.kappa.reshape((1,self.kappa.shape[0],1,1))
        lamda = self.lamda.reshape((1,self.lamda.shape[0],1,1))
        eps = self.alpha/(2**self.precision.get_bits()-1)
        return pact_quantize_inference((x.data[:] - lamda) / kappa, eps, self.alpha)

class PACT_QuantizedBatchNormNd(torch.nn.Module):
    r"""PACT-quantized N-dimensional BatchNorm.

    Implements a :py:class:`torch.nn.Module` to implement a :py:class:`torch.nn.BatchNorm2d` or :py:class:`torch.nn.BatchNorm1d` with quantized parameters.

    """

    def __init__(self, precision=None, kappa=None, lamda=None, nb_channels=1, statistics_only=False, dimensions=2, **kwargs):
        r"""Constructor.

        :param precision_kappa: instance defining the current quantization level (default `None`).
        :type  precision_kappa: :py:class:`nemo.precision.Precision`
        :param precision_lamda: instance defining the current quantization level (default `None`).
        :type  precision_lamda: :py:class:`nemo.precision.Precision`
        :param kappa: the value of :math:`\kappa`.
        :type  kappa: `torch.Tensor` or float
        :param lamda: the value of :math:`\lambda`.
        :type  lamda: `torch.Tensor` or float
        :param nb_channels: number of channels to batch-normalize.
        :type  nb_channels: int
        :param statistics_only: initialization value of `statistics_only` member.
        :type  statistics_only: bool
        :param dimensions: number of BatchNorm dimensions (default 2).
        :type  dimensions: int

        """

        super(PACT_QuantizedBatchNormNd, self).__init__()
        if precision is None:
            self.precision_kappa = Precision(bits=DEFAULT_QBATCHNORM_PREC)
            self.precision_lamda = Precision(bits=DEFAULT_QBATCHNORM_PREC)
        else:
            self.precision_kappa = Precision(bits=precision.get_bits())
            self.precision_lamda = Precision(bits=precision.get_bits())
        device = 'cuda' if torch.cuda.is_available() else 'cpu'

        if dimensions == 2:
            param_shape = lambda n : (1, n, 1, 1)
        elif dimensions == 1:
            param_shape = lambda n : (n,)

        if kappa is None:
            self.kappa = torch.nn.Parameter(torch.zeros(*param_shape(nb_channels)).to(device), requires_grad=False)
        else:
            self.kappa = torch.nn.Parameter(kappa.to(device).reshape(param_shape(kappa.shape[0])), requires_grad=False)
        if lamda is None:
            self.lamda = torch.nn.Parameter(torch.zeros(*param_shape(nb_channels)).to(device), requires_grad=False)
        else:
            self.lamda = torch.nn.Parameter(lamda.to(device).reshape(param_shape(lamda.shape[0])), requires_grad=False) 

        self.statistics_only = statistics_only

        self.min = torch.nn.Parameter(torch.zeros(1).to(device), requires_grad=False)
        self.max = torch.nn.Parameter(torch.zeros(1).to(device), requires_grad=False)

        self.onnx_qd_output = False
        self.eps_kappa = None
        self.eps_lamda = None
        self.eps_in = None
        self.eps_lamda_min = 1e-6
        self.hardened = False
        
    def harden_weights(self):
        r"""Replaces the current value of weight tensors (full-precision, quantized on forward-prop) with the quantized value.

        """

        if not self.hardened:
            kappa_clip = self.kappa.abs().max()
            lamda_clip = self.lamda.abs().max()
            self.eps_lamda = self.eps_kappa * self.eps_in
            self.kappa.data[:] = pact_quantize_signed_inference(self.kappa, self.eps_kappa, kappa_clip)
            self.lamda.data[:] = pact_quantize_signed_inference(self.lamda, self.eps_lamda, lamda_clip)
            self.hardened = True

    def get_output_eps(self, eps_in):
        r"""Get the output quantum (:math:`\varepsilon`) given the input one.

        :param eps_in: input quantum :math:`\varepsilon_{in}`.
        :type  eps_in: :py:class:`torch.Tensor`
        :return: output quantum :math:`\varepsilon_{out}`.
        :rtype:  :py:class:`torch.Tensor`

        """

        if self.eps_kappa is None or self.eps_lamda is None:
            kappa_int = self.kappa.abs().max()
            eps_kappa = 2*kappa_int/(2**self.precision_kappa.get_bits()-1)
        else:
            eps_kappa = self.eps_kappa
        self.eps_kappa = eps_kappa
        self.eps_lamda = (eps_kappa * eps_in).clone().detach()
        return eps_kappa * eps_in

    def forward(self, x):
        r"""Forward-prop function for PACT-quantized batch-norm.
        
        In forward propagation, the function is defined as 
        
        .. math::
            \mathbf{y} = \kappa\cdot\varphi + \lambda

        This layer is not meant to be trained directly, and as such using it in backward is not recommended.

        :param x: input activations tensor.
        :type  x: :py:class:`torch.Tensor`
        
        :return: output activations tensor.
        :rtype:  :py:class:`torch.Tensor`

        """

        kappa = self.kappa
        lamda = self.lamda
        if self.onnx_qd_output: # this representation is *only* used to get a nice ONNX output, should *not* be used for actual deployment!
            if self.eps_kappa is None:
                with torch.no_grad():
                    kappa_int = self.kappa.abs().max()
                    eps_kappa = 2*kappa_int/(2**self.precision_kappa.get_bits()-1)
            else:
                eps_kappa = self.eps_kappa
            if self.eps_lamda is None:
                with torch.no_grad():
                    lamda_int = self.lamda.abs().max()
                    eps_lamda = 2*lamda_int/(2**self.precision_lamda.get_bits()-1)
            else:
                eps_lamda = self.eps_lamda
            y = kappa * x + lamda
            return y
        if not self.statistics_only and not self.hardened:
            with torch.no_grad():
                kappa_int = self.kappa.abs().max()
                lamda_int = self.lamda.abs().max()
                eps_kappa = 2*kappa_int/(2**self.precision_kappa.get_bits()-1)
                eps_lamda = 2*lamda_int/(2**self.precision_lamda.get_bits()-1)
                kappa_int = torch.floor(kappa_int / eps_kappa) * eps_kappa
                lamda_int = torch.floor(lamda_int / eps_lamda) * eps_lamda
            kappa = pact_quantize_signed_inference(kappa, eps_kappa, kappa_int)
            lamda = pact_quantize_signed_inference(lamda, eps_lamda, lamda_int)
        elif not self.statistics_only and self.hardened:
            eps_kappa = self.eps_kappa
            eps_lamda = self.eps_lamda
        out = kappa*x + lamda
        if self.statistics_only:
            with torch.no_grad():
                self.min[:] = min(self.min.item(), out.min())
                self.max[:] = max(self.max.item(), out.max())
        return out

class PACT_IntegerBatchNormNd(torch.nn.Module):
    r"""Integer N-dimensional BatchNorm.

    Implements a :py:class:`torch.nn.Module` to implement a :py:class:`torch.nn.BatchNorm2d` or :py:class:`torch.nn.BatchNorm1d` with integer parameters.

    """

    def __init__(self, kappa, lamda, eps_in, eps_kappa, eps_lamda):
        r"""Constructor.

        :param kappa: the value of :math:`\kappa`.
        :type  kappa: `torch.Tensor` or float
        :param lamda: the value of :math:`\lambda`.
        :type  lamda: `torch.Tensor` or float
        :param eps_in: the value of :math:`\varepsilon_{in}`.
        :type  eps_in: `torch.Tensor` or float
        :param eps_kappa: the value of :math:`\varepsilon_\kappa`.
        :type  eps_kappa: `torch.Tensor` or float
        :param eps_lamda: the value of :math:`\varepsilon_\lambda`.
        :type  eps_lamda: `torch.Tensor` or float

        """

        super(PACT_IntegerBatchNormNd, self).__init__()
        self.kappa = kappa
        self.lamda = lamda
        self.eps_kappa = eps_kappa
        self.eps_lamda = eps_lamda
        self.eps_in = eps_in
        
    def integerize_weights(self, **kwargs):
        r"""Replaces the current value of weight tensors with the integer weights (i.e., the weight's quantized image).

        """

        self.kappa.data[:] = self.kappa / self.eps_kappa
        self.lamda.data[:] = self.lamda / self.eps_lamda

    def get_output_eps(self, eps_in):
        r"""Get the output quantum (:math:`\varepsilon`) given the input one.

        :param eps_in: input quantum :math:`\varepsilon_{in}`.
        :type  eps_in: :py:class:`torch.Tensor`
        :return: output quantum :math:`\varepsilon_{out}`.
        :rtype:  :py:class:`torch.Tensor`

        """

        return self.eps_kappa * eps_in

    def forward(self, x):
        r"""Forward-prop function for integer batch-norm.
        
        In forward propagation, the function is defined as 
        
        .. math::
            q_\mathbf{y}(\mathbf{y}) = {q}_\kappa(\kappa)\cdot{q}_\varphi(\varphi) + {q}_{\kappa\varphi}(\lambda)
        
        where :math:`q_{\kappa\varphi}(\lambda)` is obtained by requantization from :math:`\varepsilon_\lambda` to
        :math:`\varepsilon_\kappa\cdot\varepsilon_\varphi`. 

        This layer is not meant to be trained directly, and as such using it in backward is not recommended.

        :param x: input activations tensor.
        :type  x: :py:class:`torch.Tensor`
        
        :return: output activations tensor.
        :rtype:  :py:class:`torch.Tensor`

        """

        return self.kappa*x + self.lamda

class PACT_Identity(torch.nn.Module):
    r"""Identity module.

    Implements a :py:class:`torch.nn.Module` returning its own input unchanged.

    """

    def __init__(self):
        r"""Constructor.

        """

        super(PACT_Identity, self).__init__()
        
    def forward(self, x):
        r"""Forward-prop function for identity.
        
        :param x: input activations tensor.
        :type  x: :py:class:`torch.Tensor`
        
        :return: output activations tensor.
        :rtype:  :py:class:`torch.Tensor`

        """
        return x

class PACT_Conv2d(torch.nn.Conv2d):
    r"""PACT (PArametrized Clipping acTivation) 2d convolution, extending :py:class:`torch.nn.Conv2d`.

    Implements a :py:class:`torch.nn.Module` to implement PACT-like convolution. It uses PACT-like quantized weights and can be configured to quantize also input activations.
    It assumes full-precision storage of partial results.

    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        quantize_x=False,
        quantize_W=True,
        quantize_y=False,
        W_precision=None,
        x_precision=None,
        alpha=1.,
        quant_asymm=True,
        **kwargs
    ):
        r"""Constructor. Supports all arguments supported by :py:class:`torch.nn.Conv2d` plus additional ones.

        :param quantize_x: if True, quantize input activations (default False).
        :type  quantize_x: bool
        :param quantize_W: if True, quantize weights (default True).
        :type  quantize_W: bool
        :param x_precision: precision to be used for quantization of input activations (default None).
        :type  x_precision: :py:class:`nemo.precision.Precision`
        :param W_precision: precision to be used for quantization of weights (default None).
        :type  W_precision: :py:class:`nemo.precision.Precision`
        :param alpha: the value of :math:`\alpha` (default 1.0).
        :type  alpha: `torch.Tensor` or float
        :param quant_asymm: use asymmetric quantization for weights (default `True`).
        :type  quant_asymm: bool
        """
        if W_precision is None:
            self.W_precision = Precision()
        else:
            self.W_precision = Precision(bits=W_precision.get_bits())
        if x_precision is None:
            self.x_precision = Precision()
        else:
            self.x_precision = Precision(bits=x_precision.get_bits())

        self.quantize_x = quantize_x
        self.quantize_W = quantize_W

        super(PACT_Conv2d, self).__init__(in_channels, out_channels, kernel_size, **kwargs)
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.W_alpha = torch.nn.Parameter(torch.Tensor((alpha,)).to(device))
        if quant_asymm:
            self.W_beta  = torch.nn.Parameter(torch.Tensor((alpha,)).to(device))

        self.x_alpha = torch.nn.Parameter(torch.Tensor((2.0,)).to(device))
        self.weight.data.uniform_(-1., 1.)

        self.quant_asymm = quant_asymm

        self.train_loop = True
        self.deployment = False
        self.train_loop_oldprec = None

        self.padding_value = 0
        self.hardened = False
        self.integerized = False
        self.eps_out_static = None

    def reset_alpha_weights(self, use_method='max', nb_std=5., verbose=False, dyn_range_bins=1024, dyn_range_cutoff=0., **kwargs):
        r"""Resets :math:`\alpha` and :math:`\beta` parameters for weights.

        """

        if not self.quant_asymm:
            self.W_alpha.data[0] = self.weight.data.abs().max()
        elif use_method=='max':
            self.W_alpha.data[0] = -self.weight.data.min()
            self.W_beta.data [0] =  self.weight.data.max()
        elif use_method=='std':
            self.W_alpha.data[0] = -self.weight.data.mean() + nb_std*self.weight.data.std()
            self.W_beta.data[0]  =  self.weight.data.mean() + nb_std*self.weight.data.std()
        elif use_method=='dyn_range':
            import scipy.stats
            alpha_n = self.weight.data.min()
            beta    = self.weight.data.max()
            x = np.linspace(alpha_n.cpu().detach().numpy(), beta.cpu().detach().numpy(), dyn_range_bins)
            res = scipy.stats.cumfreq(self.weight.data.cpu().detach().numpy(), dyn_range_bins, defaultreallimits=(alpha_n.cpu().detach().numpy(), beta.cpu().detach().numpy()))
            yh = res.cumcount / res.cumcount[-1]
            if not (yh<dyn_range_cutoff).any() and not (yh>1-dyn_range_cutoff).any():
                self.W_alpha.data[0] = torch.as_tensor(-x.min(), device=self.W_alpha.data.device)
                self.W_beta.data[0]  = torch.as_tensor(x.max(), device=self.W_beta.data.device)
            elif not (yh<dyn_range_cutoff).any():
                self.W_alpha.data[0] = torch.as_tensor(-x[yh>1-dyn_range_cutoff].min(), device=self.W_alpha.data.device)
                self.W_beta.data[0]  = torch.as_tensor(x.max(), device=self.W_beta.data.device)
            elif not (yh>1-dyn_range_cutoff).any():
                self.W_alpha.data[0] = torch.as_tensor(-x.min(), device=self.W_alpha.data.device)
                self.W_beta.data[0]  = torch.as_tensor(x[yh<dyn_range_cutoff].max(), device=self.W_beta.data.device)
            else:
                self.W_alpha.data[0] = torch.as_tensor(-x[yh>1-dyn_range_cutoff].min(), device=self.W_alpha.data.device)
                self.W_beta.data[0]  = torch.as_tensor(x[yh<dyn_range_cutoff].max(), device=self.W_beta.data.device)
            if self.W_alpha < 0:
                self.W_alpha.data[:] = -self.W_alpha.data[:]
            if self.W_beta < 0:
                self.W_beta.data[:] = -self.W_beta.data[:]
            assert (self.W_alpha >= 0).all()
            assert (self.W_beta >= 0).all()

        if verbose:
            logging.info("[Quant] W_alpha = %.5f vs W_min = %.5f" % (self.W_alpha.data[0], self.weight.min()))
            logging.info("[Quant] W_beta  = %.5f vs W_max = %.5f" % (self.W_beta.data[0], self.weight.max()))

    def harden_weights(self):
        r"""Replaces the current value of weight tensors (full-precision, quantized on forward-prop) with the quantized value.

        """

        if not self.hardened:
            # here, clipping parameters are also quantized in order to cope with the PACT variant utilized here.
            # in this way, the ID version will be able to use only an integer displacement or none at all if
            # symmetric weights are used
            if self.quant_asymm:
                eps = (self.W_beta+self.W_alpha)/(2.0**(self.W_precision.get_bits())-1)
                self.weight.data = pact_quantize_asymm_inference(self.weight, eps, torch.ceil(self.W_alpha/eps)*eps, torch.floor(self.W_beta/eps)*eps, train_loop=False, train_loop_oldprec=self.train_loop_oldprec)
                self.eps_static = eps
            else: 
                eps = (2*self.W_alpha)/(2.0**(self.W_precision.get_bits())-1)
                self.weight.data = pact_quantize_signed_inference(self.weight, eps, self.W_alpha)
            self.hardened = True

    def integerize_weights(self, **kwargs):
        r"""Replaces the current value of weight tensors with the integer weights (i.e., the weight's quantized image).

        """
      
        if not self.integerized:
            if self.quant_asymm:
                eps = self.eps_static
                self.weight.data = self.weight.data/self.eps_static
            else:
                eps = 2*self.W_alpha/(2.0**(self.W_precision.get_bits())-1)
                self.weight.data = pact_quantize_signed_inference(self.weight, eps, self.W_alpha) / eps
            self.integerized = True

    def prune_weights(self, threshold=0.1, eps=2**-9.):
        r"""Prunes the weights of the layer.

        The pruning is performed channel-wise by replacing all weights that are "near" to the channel-wise mean with the mean itself.
        "Near" weights are those that differ with the mean by less than `threshold` times the standard deviation.

        :param threshold: threshold used for pruning (default 0.1).
        :type  threshold: float
        :param eps: parameter used to compare the difference with 0 (default 2^-9).
        :type  eps: float

        """

        logging.info("[Pruning] tau=%.1e", threshold)
        stdev_per_chan = self.weight.data.std ((2,3), keepdim=True)
        mean_per_chan  = self.weight.data.mean((2,3), keepdim=True) # + eps
        self.weight.data = torch.where((self.weight.data - mean_per_chan).abs() < stdev_per_chan*threshold, mean_per_chan, self.weight.data)
        wc = self.weight.data.clone().detach().to('cpu').numpy().flatten()
        logging.info("[Pruning] Pruned %d" % np.count_nonzero(wc < eps))
        return np.count_nonzero(wc < eps)

    def get_output_eps(self, eps_in):
        r"""Get the output quantum (:math:`\varepsilon`) given the input one.

        :param eps_in: input quantum :math:`\varepsilon_{in}`.
        :type  eps_in: :py:class:`torch.Tensor`
        :return: output quantum :math:`\varepsilon_{out}`.
        :rtype:  :py:class:`torch.Tensor`

        """

        if self.eps_out_static is None:
            if self.quant_asymm:
                eps_W = (self.W_beta+self.W_alpha)/(2.0**(self.W_precision.get_bits())-1)
            else:
                eps_W = 2*self.W_alpha/(2.0**(self.W_precision.get_bits())-1)
            self.eps_out_static = eps_W * eps_in
        return self.eps_out_static

    def forward(self, input):
        r"""Forward-prop function for PACT-quantized 2d-convolution.

        :param input: input activations tensor.
        :type  input: `torch.Tensor`

        """

        if self.training and self.quantize_x and not self.deployment:
            x_quant = pact_quantize_signed(input, self.x_alpha/(2.0**(self.x_precision.get_bits())-1), self.x_alpha)
        else:
            x_quant = input
        if self.training and self.quantize_W and not self.deployment:
            if self.quant_asymm:
                W_quant = pact_quantize_asymm(self.weight, (self.W_beta+self.W_alpha)/(2.0**(self.W_precision.get_bits())-1), self.W_alpha, self.W_beta)
            else:
                W_quant = pact_quantize_signed(self.weight, 2*self.W_alpha/(2.0**(self.W_precision.get_bits())-1), self.W_alpha)
        elif self.quantize_W and not self.deployment:
            if self.quant_asymm:
                eps = (self.W_beta+self.W_alpha)/(2.0**(self.W_precision.get_bits())-1)
                W_quant = pact_quantize_asymm_inference(self.weight, eps, torch.ceil(self.W_alpha/eps)*eps, torch.floor(self.W_beta/eps)*eps, train_loop=self.train_loop, train_loop_oldprec=self.train_loop_oldprec)
            else:
                W_quant = pact_quantize_signed_inference(self.weight, 2*self.W_alpha/(2.0**(self.W_precision.get_bits())-1), self.W_alpha)
        else:
            W_quant = self.weight
        # if input bias is present, padding should be performed using the input bias as padding value instead of 0
        if self.deployment and self.padding is not None and self.bias is not None:
            if type(self.padding) is not tuple and type(self.padding) is not list:
                pad = (self.padding, self.padding, self.padding, self.padding)
            elif len(self.padding) == 2:
                pad = (*self.padding, *self.padding)
            else:
                pad = self.padding
            x_quant = torch.nn.functional.pad(x_quant, pad, 'constant', self.padding_value)
        y = torch.nn.functional.conv2d(
            x_quant,
            W_quant,
            self.bias, # typically nil
            self.stride,
            self.padding if not self.deployment or self.bias is None else 0,
            self.dilation,
            self.groups
        )
        if not self.training and self.quantize_W:
            del W_quant
        # y is returned non-quantized, as it is assumed to be quantized after BN
        return y

class PACT_Conv1d(torch.nn.Conv1d):
    r"""PACT (PArametrized Clipping acTivation) 1d convolution, extending :py:class:`torch.nn.Conv1d`.

    Implements a :py:class:`torch.nn.Module` to implement PACT-like convolution. It uses PACT-like quantized weights and can be configured to quantize also input activations.
    It assumes full-precision storage of partial results.

    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        quantize_x=False,
        quantize_W=True,
        W_precision=None,
        x_precision=None,
        alpha=1.,
        quant_asymm=True,
        **kwargs
    ):
        r"""Constructor. Supports all arguments supported by :py:class:`torch.nn.Conv2d` plus additional ones.

        :param quantize_x: if True, quantize input activations (default False).
        :type  quantize_x: bool
        :param quantize_W: if True, quantize weights (default True).
        :type  quantize_W: bool
        :param x_precision: precision to be used for quantization of input activations (default None).
        :type  x_precision: :py:class:`nemo.precision.Precision`
        :param W_precision: precision to be used for quantization of weights (default None).
        :type  W_precision: :py:class:`nemo.precision.Precision`
        :param alpha: the value of :math:`\alpha` (default 1.0).
        :type  alpha: `torch.Tensor` or float
        :param quant_asymm: use asymmetric quantization for weights (default `True`).
        :type  quant_asymm: bool
        """

        if W_precision is None:
            self.W_precision = Precision()
        else:
            self.W_precision = Precision(bits=W_precision.get_bits())
        if x_precision is None:
            self.x_precision = Precision()
        else:
            self.x_precision = Precision(bits=x_precision.get_bits())

        self.quantize_x = quantize_x
        self.quantize_W = quantize_W
        super(PACT_Conv1d, self).__init__(in_channels, out_channels, kernel_size, **kwargs)
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.W_alpha = torch.nn.Parameter(torch.Tensor((alpha,)).to(device))
        if quant_asymm:
            self.W_beta  = torch.nn.Parameter(torch.Tensor((alpha,)).to(device))

        self.x_alpha = torch.nn.Parameter(torch.Tensor((2.0,)).to(device))
        self.weight.data.uniform_(-1., 1.)

        # FIXME to implement: fix alpha,beta scaling factors for "beautiful" quantization in INT4,6,8

        self.quant_asymm = quant_asymm

        self.train_loop = True
        self.deployment = False
        self.train_loop_oldprec = None
        self.hardened = False

    def reset_alpha_weights(self, use_max=True, nb_std=5., verbose=False, **kwargs):
        r"""Resets :math:`\alpha` and :math:`\beta` parameters for weights.

        """

        if not self.quant_asymm:
            self.W_alpha.data[0] = self.weight.data.abs().max()
        elif use_max:
            self.W_alpha.data[0] = -self.weight.data.min()
            self.W_beta.data [0] =  self.weight.data.max()
        else:
            self.W_alpha.data[0] = -self.weight.data.mean() + nb_std*self.weight.data.std()
            self.W_beta.data[0]  =  self.weight.data.mean() + nb_std*self.weight.data.std()
        if verbose:
            logging.info("[Quant] W_alpha = %.5f" % self.W_alpha.data[0])
            logging.info("[Quant] W_beta  = %.5f" % self.W_beta.data[0])


    def harden_weights(self):
        r"""Replaces the current value of weight tensors (full-precision, quantized on forward-prop) with the quantized value.

        """

        if not self.hardened:
            # here, clipping parameters are also quantized in order to cope with the PACT variant utilized here.
            # in this way, the ID version will be able to use only an integer displacement or none at all if
            # symmetric weights are used
            if self.quant_asymm:
                self.reset_alpha_weights()
                eps = (self.W_beta+self.W_alpha)/(2.0**(self.W_precision.get_bits())-1)
                self.weight.data = pact_quantize_asymm_inference(self.weight, eps, torch.ceil(self.W_alpha/eps)*eps, torch.floor(self.W_beta/eps)*eps, train_loop=False, train_loop_oldprec=self.train_loop_oldprec)
                self.reset_alpha_weights()
            else: 
                eps = (2*self.W_alpha)/(2.0**(self.W_precision.get_bits())-1)
                self.weight.data = pact_quantize_signed_inference(self.weight, eps, self.W_alpha)
            self.hardened = True

    def integerize_weights(self, **kwargs):
        r"""Replaces the current value of weight tensors with the integer weights (i.e., the weight's quantized image).

        """

        if self.quant_asymm:
            eps = (self.W_beta+self.W_alpha)/(2.0**(self.W_precision.get_bits())-1)
            self.weight.data = pact_quantize_asymm_inference(self.weight, eps, self.W_alpha, self.W_beta, train_loop=False) / eps
        else:
            eps = 2*self.W_alpha/(2.0**(self.W_precision.get_bits())-1)
            self.weight.data = pact_quantize_signed_inference(self.weight, eps, self.W_alpha) / eps

    def prune_weights(self, threshold=0.1, eps=2**-9.):
        # logging.info("[Pruning] tau=%.1e", threshold)
        # stdev_per_chan = self.weight.data.std ((2), keepdim=True)
        # mean_per_chan  = self.weight.data.mean((2), keepdim=True) # + eps
        # self.weight.data = torch.where((self.weight.data - mean_per_chan).abs() < stdev_per_chan*threshold, mean_per_chan, self.weight.data)
        # wc = self.weight.data.clone().detach().to('cpu').numpy().flatten()
        # logging.info("[Pruning] Pruned %d" % np.count_nonzero(wc < eps))
        # return np.count_nonzero(wc < eps)
        return 0

    def get_output_eps(self, eps_in):
        r"""Get the output quantum (:math:`\varepsilon`) given the input one.

        :param eps_in: input quantum :math:`\varepsilon_{in}`.
        :type  eps_in: :py:class:`torch.Tensor`
        :return: output quantum :math:`\varepsilon_{out}`.
        :rtype:  :py:class:`torch.Tensor`

        """
        if self.quant_asymm:
            eps_W = (self.W_beta+self.W_alpha)/(2.0**(self.W_precision.get_bits())-1)
        else:
            eps_W = 2*self.W_alpha/(2.0**(self.W_precision.get_bits())-1)
        return eps_W * eps_in

    def forward(self, input):
        r"""Forward-prop function for PACT-quantized 1d-convolution.

        :param input: input activations tensor.
        :type  input: `torch.Tensor`

        """

        if self.training and self.quantize_W and not self.deployment:
            if self.quant_asymm:
                W_quant = pact_quantize_asymm(self.weight, (self.W_beta+self.W_alpha)/(2.0**(self.W_precision.get_bits())-1), self.W_alpha, self.W_beta)
            else:
                W_quant = pact_quantize_signed(self.weight, 2*self.W_alpha/(2.0**(self.W_precision.get_bits())-1), self.W_alpha)
        elif self.quantize_W and not self.deployment:
            if self.quant_asymm:
                W_quant = pact_quantize_asymm_inference(self.weight, (self.W_beta+self.W_alpha)/(2.0**(self.W_precision.get_bits())-1), self.W_alpha, self.W_beta, train_loop=self.train_loop)
            else:
                W_quant = pact_quantize_signed_inference(self.weight, 2*self.W_alpha/(2.0**(self.W_precision.get_bits())-1), self.W_alpha)
        else:
            W_quant = self.weight
        if self.padding_mode == 'circular':
            expanded_padding = ((self.padding[0] + 1) // 2, self.padding[0] // 2)
            return torch.nn.functional.conv1d(torch.nn.functional.pad(input, expanded_padding, mode='circular'),
                            W_quant, self.bias, self.stride,
                            _single(0), self.dilation, self.groups)
        return torch.nn.functional.conv1d(input, W_quant, self.bias, self.stride,
                        self.padding, self.dilation, self.groups)

class PACT_Linear(torch.nn.Linear):
    r"""PACT (PArametrized Clipping acTivation) fully connected layer, extending :py:class:`torch.nn.Linear`.

    Implements a :py:class:`torch.nn.Module` to implement PACT-like matrix multiplication. It uses PACT-like quantized weights and can be configured to quantize also input activations.
    It assumes full-precision storage of partial results.

    """

    def __init__(
        self,
        in_features,
        out_features,
        quantize_x=False,
        quantize_W=True,
        W_precision=None,
        x_precision=None,
        alpha=1.,
        quant_asymm=True,
        quant_pc=False,
        **kwargs
    ):
        if W_precision is None:
            self.W_precision = Precision()
        else:
            self.W_precision = Precision(bits=W_precision.get_bits())
        if x_precision is None:
            self.x_precision = Precision()
        else:
            self.x_precision = Precision(bits=x_precision.get_bits())

        self.quantize_x = quantize_x
        self.quantize_W = quantize_W
        super(PACT_Linear, self).__init__(in_features, out_features, **kwargs)
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.W_alpha = torch.nn.Parameter(torch.Tensor((alpha,)).to(device))
        if quant_asymm:
            self.W_beta  = torch.nn.Parameter(torch.Tensor((alpha,)).to(device))
        self.x_alpha = torch.nn.Parameter(torch.Tensor((2.0,)).to(device))
        self.weight.data.uniform_(-1., 1.)

        self.quant_asymm = quant_asymm
        self.quant_pc    = quant_pc
        
        self.train_loop = True
        self.deployment = False
        self.train_loop_oldprec = None
        self.hardened = False

    def reset_alpha_weights(self, use_max=True, nb_std=5., verbose=False, **kwargs):
        r"""Resets :math:`\alpha` and :math:`\beta` parameters for weights.

        """

        if not self.quant_asymm:
            self.W_alpha.data[0] = self.weight.data.abs().max()
        elif use_max:
            self.W_alpha.data[0] = -self.weight.data.min()
            self.W_beta.data [0] =  self.weight.data.max()
        else:
            self.W_alpha.data[0] = -self.weight.data.mean() + nb_std*self.weight.data.std()
            self.W_beta.data[0]  =  self.weight.data.mean() + nb_std*self.weight.data.std()
        if verbose:
            logging.info("[Quant] W_alpha = %.5f" % self.W_alpha.data[0])
            logging.info("[Quant] W_beta  = %.5f" % self.W_beta.data[0])

    def harden_weights(self):
        r"""Replaces the current value of weight tensors (full-precision, quantized on forward-prop) with the quantized value.

        """

        if not self.hardened:
            # here, clipping parameters are also quantized in order to cope with the PACT variant utilized here.
            # in this way, the ID version will be able to use only an integer displacement or none at all if
            # symmetric weights are used
            if self.quant_asymm:
                self.reset_alpha_weights()
                eps = (self.W_beta+self.W_alpha)/(2.0**(self.W_precision.get_bits())-1)
                self.weight.data = pact_quantize_asymm_inference(self.weight, eps, torch.ceil(self.W_alpha/eps)*eps, torch.floor(self.W_beta/eps)*eps, train_loop=False, train_loop_oldprec=self.train_loop_oldprec)
                self.reset_alpha_weights()
            else: 
                eps = (2*self.W_alpha)/(2.0**(self.W_precision.get_bits())-1)
                self.weight.data = pact_quantize_signed_inference(self.weight, eps, self.W_alpha)
            self.hardened = True

    def integerize_weights(self, **kwargs):
        r"""Replaces the current value of weight tensors with the integer weights (i.e., the weight's quantized image).

        """

        if self.quant_asymm:
            eps = (self.W_beta+self.W_alpha)/(2.0**(self.W_precision.get_bits())-1)
            self.weight.data = pact_quantize_asymm_inference(self.weight, eps, self.W_alpha, self.W_beta, train_loop=False) / eps
        else:
            eps = 2*self.W_alpha/(2.0**(self.W_precision.get_bits())-1)
            self.weight.data = pact_quantize_signed_inference(self.weight, eps, self.W_alpha) / eps

    def prune_weights(self, threshold=0.1, eps=2**-9.):
        r"""Prunes the weights of the layer.

        The pruning is performed by replacing all weights that are "near" to the mean of the weights with the mean itself.
        "Near" weights are those that differ with the mean by less than `threshold` times the standard deviation.

        :param threshold: threshold used for pruning (default 0.1).
        :type  threshold: float
        :param eps: parameter used to compare the difference with 0 (default 2^-9).
        :type  eps: float

        """

        logging.info("[Pruning] tau=%.1e", threshold)
        stdev_per_chan = self.weight.data.std ()
        mean_per_chan  = self.weight.data.mean() # + eps
        self.weight.data = torch.where((self.weight.data - mean_per_chan).abs() < stdev_per_chan*threshold, mean_per_chan, self.weight.data)
        wc = self.weight.data.clone().detach().to('cpu').numpy().flatten()
        logging.info("[Pruning] Pruned %d" % np.count_nonzero(wc < eps))
        return np.count_nonzero(wc < eps)

    def get_output_eps(self, eps_in):
        r"""Get the output quantum (:math:`\varepsilon`) given the input one.

        :param eps_in: input quantum :math:`\varepsilon_{in}`.
        :type  eps_in: :py:class:`torch.Tensor`
        :return: output quantum :math:`\varepsilon_{out}`.
        :rtype:  :py:class:`torch.Tensor`

        """

        if self.quant_asymm:
            eps_W = (self.W_beta+self.W_alpha)/(2.0**(self.W_precision.get_bits())-1)
        else:
            eps_W = 2*self.W_alpha/(2.0**(self.W_precision.get_bits())-1)
        return eps_W * eps_in

    def forward(self, input):
        r"""Forward-prop function for PACT-quantized linear layer.

        :param input: input activations tensor.
        :type  input: `torch.Tensor`

        """

        if self.training and self.quantize_x and not self.deployment:
            x_quant = pact_quantize_signed(input, self.x_alpha/(2.0**(self.x_precision.get_bits())-1), self.x_alpha)
        else:
            x_quant = input
        if self.training and self.quantize_W and not self.deployment:
            if self.quant_asymm:
                W_quant = pact_quantize_asymm(self.weight, (self.W_beta+self.W_alpha)/(2.0**(self.W_precision.get_bits())-1), self.W_alpha, self.W_beta)
            else:
                W_quant = pact_quantize_signed(self.weight, 2*self.W_alpha/(2.0**(self.W_precision.get_bits())-1), self.W_alpha)
        elif self.quantize_W and not self.deployment:
            if self.quant_asymm:
                W_quant = pact_quantize_asymm_inference(self.weight, (self.W_beta+self.W_alpha)/(2.0**(self.W_precision.get_bits())-1), self.W_alpha, self.W_beta, train_loop=self.train_loop)
            else:
                W_quant = pact_quantize_signed_inference(self.weight, 2*self.W_alpha/(2.0**(self.W_precision.get_bits())-1), self.W_alpha)
        else:
            W_quant = self.weight
        y = torch.nn.functional.linear(x_quant, W_quant, self.bias)
        if not self.training and self.quantize_W:
            del W_quant
        # y is returned non-quantized, as it is assumed to be quantized after BN
        return y



#
# pruning.py
# Francesco Conti <fconti@iis.ee.ethz.ch>
#
# Copyright (C) 2018-2020 ETH Zurich
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from nemo.precision import Precision
from nemo.quant.pact import *
from nemo.graph import DeployGraph
from torch.nn.modules.utils import _single,_pair
from collections import OrderedDict
import types
import logging
import numpy as np
import copy
import math
import torchvision.models
import re
from nemo.transf.common import *

def _prune_weights_pact(self, **kwargs):
    r"""Prune weights.

    """

    pruned = 0
    for n,m in self.named_modules():
        if (m.__class__.__name__ == "PACT_Conv2d" or \
            m.__class__.__name__ == "PACT_Conv1d" or \
            m.__class__.__name__ == "PACT_Linear"):
            pruned += m.prune_weights(**kwargs)
    return pruned



#
# utils.py
# Francesco Conti <fconti@iis.ee.ethz.ch>
# Alfio Di Mauro <adimauro@iis.ee.ethz.ch>
#
# Copyright (C) 2018-2020 ETH Zurich
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import logging
import os
import json
import re
import nemo
import numpy as np
from collections import OrderedDict

def precision_dict_to_json(d, filename=None):
    s = json.dumps(d, indent=4)
    if filename is not None:
        with open(filename, 'w') as f:
            f.write(s)
    else:
        return s

def precision_dict_from_json(filename):
    with open(filename, "r") as f:
        rr = json.load(f)
    return rr

def process_json(json, args):
    args = vars(args)
    regime = {}
    if args.regime is not None:
        with open(args.regime, "r") as f:
            rr = json.load(f)
        for k in rr.keys():
            try:
                regime[int(k)] = rr[k]
            except ValueError:
                regime[k] = rr[k]

def save_checkpoint(net, optimizer, epoch, acc=0.0, checkpoint_name='net_', checkpoint_suffix=''):
    checkpoint_name = checkpoint_name + checkpoint_suffix
    logging.info('Saving checkpoint %s' % checkpoint_name)
    try:
        optimizer_state = optimizer.state_dict()
    except AttributeError:
        optimizer_state = None
    try:
        precision = net.export_precision()
    except AttributeError:
        precision = None
    state = {
        'epoch': epoch + 1,
        'state_dict': net.state_dict(),
        'precision': precision,
        'acc': acc,
        'optimizer' : optimizer_state,
    }
    if not os.path.isdir('checkpoint'):
        os.mkdir('checkpoint')
    torch.save(state, './checkpoint/%s.pth' % (checkpoint_name))

def export_onnx(file_name, net, net_inner, input_shape, round_params=True, perm=None, redefine_names=False, batch_size=1, verbose=False):
    if perm is None:
        perm = lambda x : x
    pattern = re.compile('[\W_]+')
    dummy_input = perm(torch.randn(batch_size, *input_shape, device='cuda' if torch.cuda.is_available() else 'cpu'))
    net.eval()
    # rounding of parameters to avoid strange numerical errors on writeout
    if round_params:
        for param in net_inner.parameters(recurse=True):
            if param.dtype is torch.float32:
                param[:] = torch.round(param)
    if redefine_names:
        input_names  = [ 'input' ] + [ pattern.sub('_', n) for n,_ in net_inner.named_parameters() ]
        output_names = [ 'output' ]
        torch.onnx.export(net_inner, dummy_input, file_name, verbose=verbose, do_constant_folding=True, input_names=input_names, output_names=output_names, export_params=True)
    else:
        torch.onnx.export(net_inner, dummy_input, file_name, verbose=verbose, do_constant_folding=True, export_params=True)

PRECISION_RULE_KEYS_REQUIRED = {
    "for_epochs": 1,
    "for_epochs_no_abs_bound": 3,
    "delta_loss_less_than": 0.01,
    "running_avg_memory": 5,
    "delta_loss_running_std_stale": 1.5,
    "abs_loss_stale": 1.4,
    "scale_lr": True,
    "lr_scaler": 4.0,
    "divergence_abs_threshold": 1e9
}
PRECISION_RULE_KEYS_ALLOWED = [
    "custom_scaler",
    "bit_scaler",
    "bit_stop_condition"
]
 
def parse_precision_rule(rule):
    required = list(PRECISION_RULE_KEYS_REQUIRED.keys())
    allowed  = PRECISION_RULE_KEYS_ALLOWED + required
    for k in required:
        if not k in rule:
            rule[k] = PRECISION_RULE_KEYS_REQUIRED[k]
    flag = False
    for k in rule.keys():
        if not k in allowed and not k.isdigit():
            print("[ERROR] %s is not a key allowed in the relaxation rule", k)
            flag = True
    if "bit_scaler" in rule and not "W_bit_scaler" in rule:
        rule["W_bit_scaler"] = rule["bit_scaler"]
    if "bit_scaler" in rule and not "x_bit_scaler" in rule:
        rule["x_bit_scaler"] = rule["bit_scaler"]
    if "bit_stop_condition" in rule and not "W_bit_stop_condition" in rule:
        rule["W_bit_stop_condition"] = rule["bit_stop_condition"]
    if "bit_stop_condition" in rule and not "x_bit_stop_condition" in rule:
        rule["x_bit_stop_condition"] = rule["bit_stop_condition"]
    if flag:
        import sys; sys.exit(1)
    print(list(rule.keys()))
    return rule
    
# see https://github.com/sksq96/pytorch-summary
def get_summary(net, input_size, batch_size=1, device="cuda", verbose=False):
    s = ""
    mdict = {}
    for n,m in net.named_modules():
        mdict[n] = m
    def register_hook(module):
        def hook(module, input, output):
            class_name = str(module.__class__).split(".")[-1].split("'")[0]
            module_idx = len(summary)
            m_key = next(n for n,m in mdict.items() if m==module)
            summary[m_key] = OrderedDict()
            summary[m_key]["input_shape"] = list(input[0].size())
            summary[m_key]["input_shape"][0] = batch_size
            if isinstance(output, (list, tuple)):
                summary[m_key]["output_shape"] = [
                    [-1] + list(o.size())[1:] for o in output
                ]
            else:
                summary[m_key]["output_shape"] = list(output.size())
                summary[m_key]["output_shape"][0] = batch_size

            params = 0
            if hasattr(module, "weight") and hasattr(module.weight, "size"):
                try:
                    params += torch.prod(torch.LongTensor(list(module.weight.size()))) / module.group
                except AttributeError:
                    params += torch.prod(torch.LongTensor(list(module.weight.size())))
                summary[m_key]["trainable"] = module.weight.requires_grad
            if hasattr(module, "bias") and hasattr(module.bias, "size"):
                params += torch.prod(torch.LongTensor(list(module.bias.size())))
            summary[m_key]["nb_params"] = params
            
            if hasattr(module, "W_precision"):
                summary[m_key]['W_bits'] = module.W_precision.get_bits()
            
            if hasattr(module, "precision"):
                summary[m_key]['bits'] = module.precision.get_bits()

        if (
            not isinstance(module, torch.nn.Sequential)
            and not isinstance(module, torch.nn.ModuleList)
            and not (module == net)
        ):
            hooks.append(module.register_forward_hook(hook))

    device = device.lower()
    assert device in [
        "cuda",
        "cpu",
    ], "Input device is not valid, please specify 'cuda' or 'cpu'"

    if device == "cuda" and torch.cuda.is_available():
        dtype = torch.cuda.FloatTensor
    else:
        dtype = torch.FloatTensor

    # multiple inputs to the network
    if isinstance(input_size, tuple):
        input_size = [input_size]

    # batch_size of 2 for batchnorm
    x = [torch.rand(2, *in_size).type(dtype) for in_size in input_size]

    # create properties
    summary = OrderedDict()
    hooks = []

    # register hook
    net.apply(register_hook)

    # make a forward pass
    net(*x)

    # remove these hooks
    for h in hooks:
        h.remove()

    s += "----------------------------------------------------------------" + "\n"
    line_new = "{:>20}  {:>25} {:>15}".format("Layer (type)", "Output Shape", "Param #")
    s += line_new + "\n"
    s += "================================================================" + "\n"
    total_params = 0
    total_output = 0
    trainable_params = 0
    params_size = 0
    output_size = 0
    input_size = 0
    for layer in summary:
        # input_shape, output_shape, trainable, nb_params
        line_new = "{:>20}  {:>25} {:>15}".format(
            layer,
            str(summary[layer]["output_shape"]),
            "{0:,}".format(summary[layer]["nb_params"]),
        )
        total_params += summary[layer]["nb_params"]
        try:
            params_size += abs(summary[layer]["nb_params"] / (1024) * summary[layer]["W_bits"] / 8.)
        except KeyError:
            params_size += abs(summary[layer]["nb_params"] / (1024) * 32. / 8.)
        total_output += np.prod(summary[layer]["output_shape"])
        try:
            output_size = max(output_size, np.prod(summary[layer]["output_shape"]) / (1024) * summary[layer]["bits"] / 8)
        except KeyError:
            output_size = max(output_size, np.prod(summary[layer]["output_shape"]) / (1024) * 32 / 8)
        if "trainable" in summary[layer]:
            if summary[layer]["trainable"] == True:
                trainable_params += summary[layer]["nb_params"]
        s += line_new + "\n"

    s += "================================================================" + "\n"
    s += "Total params: {0:,}".format(total_params) + "\n"
    s += "Trainable params: {0:,}".format(trainable_params) + "\n"
    s += "Non-trainable params: {0:,}".format(total_params - trainable_params) + "\n"
    s += "----------------------------------------------------------------" + "\n"
    s += "Biggest activation tensor size (kB): %0.2f" % output_size + "\n"
    s += "Params size (kB): %0.2f" % params_size + "\n"
    s += "----------------------------------------------------------------" + "\n"
    if verbose:
        logging.info(s)
    return summary, s

def get_intermediate_activations(net, test_fn, *test_args, **test_kwargs):
    l = len(list(net.named_modules()))
    buffer_in  = OrderedDict([])
    buffer_out = OrderedDict([])
    hooks = OrderedDict([])
    def get_hk(n):
        def hk(module, input, output):
            buffer_in  [n] = input
            buffer_out [n] = output
        return hk
    for i,(n,l) in enumerate(net.named_modules()):
        hk = get_hk(n)
        hooks[n] = l.register_forward_hook(hk)
    ret = test_fn(*test_args, **test_kwargs)
    for n,l in net.named_modules():
        hooks[n].remove()
    return buffer_in, buffer_out, ret
        
def get_intermediate_eps(net, eps_in):
    l = len(list(net.named_modules()))
    eps = OrderedDict([])
    for i,(n,l) in enumerate(net.named_modules()):
        eps[n] = net.get_eps_at(n, eps_in)
    return eps

def get_integer_activations(buf, eps, net=None):
    if type(eps) is float and net is None:
        return buf
    elif type(eps) is float:
        eps_in = eps
        eps = OrderedDict([])
        for n,m in net.named_modules():
            try:
                eps[n] = m.get_output_eps(eps_in)
            except AttributeError:
                pass
    buf_ = OrderedDict([])
    for n in buf.keys():
        b = buf.get(n, None)
        e = eps.get(n, None)
        if b is None or e is None:
            continue
        if type(buf[n]) is tuple or type(buf[n]) is list:
            buf_[n] = []
            for b in buf[n]:
                buf_[n].append((b / eps[n]).floor()) # FIXME
        else:
            buf_[n] = (buf[n] / eps[n]).floor()
    return buf_
    



#
# transform.py
# Francesco Conti <fconti@iis.ee.ethz.ch>
# Alfio Di Mauro <adimauro@iis.ee.ethz.ch>
#
# Copyright (C) 2018-2020 ETH Zurich
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from nemo.precision import Precision
from nemo.quant.pact import *
from nemo.graph import DeployGraph
from torch.nn.modules.utils import _single,_pair
from collections import OrderedDict
import types
import logging
import numpy as np
import copy
import math
import torchvision.models
import re
import nemo
from nemo.transf.common import *
from nemo.transf.bias import *
from nemo.transf.bn import *
from nemo.transf.deploy import *
from nemo.transf.equalize import *
from nemo.transf.export import *
from nemo.transf.pruning import *
from nemo.transf.statistics import *
from nemo.transf.utils import *

def quantize_pact(module, W_bits=4, x_bits=4, dummy_input=None, remove_dropout=False, **kwargs):
    r"""Takes a PyTorch module and makes it quantization-aware with PACT, recursively.

    The function follows recursively the data structures containing PyTorch layers (typically as hierarchical lists, e.g.
    block-level :py:class:`torch.nn.Sequential` for networks like ResNet).
    It performs two main kinds of replacements:
    - linear layers like :py:class:`torch.nn.Conv2d`, :py:class:`torch.nn.Conv1d`, :py:class:`torch.nn.Linear` are replaced
    with quantization-aware versions (:py:class:`nemo.quant.pact.PACT_Conv2d`, :py:class:`nemo.quant.pact.PACT_Conv1d`,
    :py:class:`nemo.quant.pact.PACT_Linear`). By default, these layers quantize weights but not (input) activations.
    - activation layers like :py:class:`torch.nn.ReLU`, :py:class:`torch.nn.ReLU6` are replaced
    with a special quantized activation, :py:class:`nemo.quant.pact.PACT_Act` that performs both clipping and quantization.

    The returned layer exposes a series of methods typical of quantization-aware models:
    - `export_precision`, exports the precision of current layers as a dictionary.
    - `change_precision`, changes the target precision of the model.
    - `reset_alpha_weights`, resets the scaling factors for weights.
    - `get_clip_parameters`, returns scaling factor parameters of activations, that can be trained.
    - `get_nonclip_parameters`, returns all parameters except scaling factor parameters of activations.
    - `harden_weights`, hardens the quantized value of weights into their underlying float representation.
    - `prune_weights`, replaces a portion of the weights in the model with 0's.
    - `set_statistics_act`, setup activation layers in statistics collection mode (they run as normal ReLUs collecting
    the maximum value of activations, to calibrate the scaling parameters).
    - `get_statistics_act`, get collected activation layer statistics.
    - `unset_statistics_act`, setup activation layers to act normally as quantization-aware layers.
    - `reset_alpha_act`, uses the collected activation layer statistics to recalibrate the scaling parameters. 
    
    :param module: module to be transformed to use PACT quantization (typically, a container like :py:class:`torch.nn.ModuleList`).
    :type  module: `torch.nn.Module`
    
    :param W_bits: target precision for weights.
    :type  W_bits: float
    
    :param x_bits: target precision for activations.
    :type  x_bits: float
    
    :param dummy_input: dummy input tensor (default None). Used to derive an adjacency map by tracing
    :type  dummy_input: `torch.Tensor`

    :param remove_dropout: if True, removes dropout layers before graph construction.
    :type  remove_dropout: bool
    
    :return: The quantization-aware module.
    :rtype:  same as `module`

    """
    # if given a dummy input, get an adjacency map of the module and other useful things
    module.eval()
    if remove_dropout:
        module = nemo.transform.dropout_to_identity(module)
    if dummy_input is not None:
        module.graph = DeployGraph(module, dummy_input=dummy_input)
    else:
        module.graph = None
    module.stage = 'fq'
    module = _hier_quantizer_pact(module, module.graph, **kwargs)
    if hasattr(module, 'graph'):
        if module.graph is not None:
            module.graph.rebuild_module_dict()
    module.add_input_bias              = types.MethodType(nemo.transf.bias._add_input_bias_pact, module)
    module.remove_bias                 = types.MethodType(nemo.transf.bias._remove_bias_pact, module)
    module.remove_input_bias           = types.MethodType(nemo.transf.bias._remove_input_bias_pact, module) 
    module.freeze_bn                   = types.MethodType(nemo.transf.bn._freeze_bn, module)
    module.unfreeze_bn                 = types.MethodType(nemo.transf.bn._unfreeze_bn, module)
    module.absorb_affine_bn            = types.MethodType(nemo.transf.bn._absorb_affine_bn, module)
    module.prime_affine_bn             = types.MethodType(nemo.transf.bn._prime_affine_bn, module)
    module.fold_bn                     = types.MethodType(nemo.transf.bn._fold_bn_pact, module)
    module.fold_bn_withinv             = types.MethodType(nemo.transf.bn._fold_bn_pact, module)
    module.fold_thresholds             = types.MethodType(nemo.transf.bn._threshold_folding_pact, module)
    module.prune_empty_bn              = types.MethodType(nemo.transf.bn._prune_empty_bn_pact, module)
    module.calibrate_bn                = types.MethodType(nemo.transf.bn._calibrate_bn_pact, module)
    module.get_eps_at                  = types.MethodType(nemo.transf.deploy._get_eps_at, module)
    module.set_eps_in                  = types.MethodType(nemo.transf.deploy._set_eps_in_pact, module)
    module.round_weights               = types.MethodType(nemo.transf.deploy._round_weights_pact, module)
    module.harden_weights              = types.MethodType(nemo.transf.deploy._harden_weights_pact, module)
    module.set_deployment              = types.MethodType(nemo.transf.deploy._set_deployment_pact, module)
    module.qd_stage                    = types.MethodType(nemo.transf.deploy._qd_stage, module)
    module.id_stage                    = types.MethodType(nemo.transf.deploy._id_stage, module)
    module.export_precision            = types.MethodType(nemo.transf.export._export_precision, module)
    module.export_weights_legacy_int16 = types.MethodType(nemo.transf.export._export_weights_legacy_int16, module)
    module.change_precision            = types.MethodType(nemo.transf.utils._change_precision_pact, module)
    module.reset_alpha_weights         = types.MethodType(nemo.transf.utils._reset_alpha_weights_pact, module)
    module.reset_alpha_act             = types.MethodType(nemo.transf.utils._reset_alpha_act_pact, module)
    module.get_clip_parameters         = types.MethodType(nemo.transf.utils._get_clip_parameters_pact, module)
    module.get_nonclip_parameters      = types.MethodType(nemo.transf.utils._get_nonclip_parameters_pact, module)
    module.set_train_loop              = types.MethodType(nemo.transf.utils._set_train_loop_pact, module)
    module.unset_train_loop            = types.MethodType(nemo.transf.utils._unset_train_loop_pact, module)
    module.prune_weights               = types.MethodType(nemo.transf.pruning._prune_weights_pact, module)
    module.equalize_weights_dfq        = types.MethodType(nemo.transf.equalize._equalize_weights_dfq_pact, module)
    module.equalize_weights_lsq        = types.MethodType(nemo.transf.equalize._equalize_weights_lsq_pact, module)
    module.equalize_weights_unfolding  = types.MethodType(nemo.transf.equalize._equalize_weights_unfolding_pact, module)
    module.statistics_act              = types.MethodType(nemo.transf.statistics._statistics_act_pact, module)
    module.set_statistics_act          = types.MethodType(nemo.transf.statistics._set_statistics_act_pact, module)
    module.get_statistics_act          = types.MethodType(nemo.transf.statistics._get_statistics_act_pact, module)
    module.unset_statistics_act        = types.MethodType(nemo.transf.statistics._unset_statistics_act_pact, module)
    module.statistics_bn               = types.MethodType(nemo.transf.statistics._statistics_bn_pact, module)
    module.set_statistics_bn           = types.MethodType(nemo.transf.statistics._set_statistics_bn_pact, module)
    module.unset_statistics_bn         = types.MethodType(nemo.transf.statistics._unset_statistics_bn_pact, module)
    module.W_precision = Precision(W_bits, None)
    module.x_precision = Precision(x_bits, None)
    return module

def _hier_replacer(module, name, replacement):
    for n,m in module.named_children():
        if n == name:
            module._modules[n] = replacement()
        elif n == name.split('.')[0]:
            module._modules[n] = _hier_replacer(m, '.'.join(name.split('.')[1:]), replacement)
    return module

def _hier_bn_to_identity(module):
    if module.__class__.__name__ == 'BatchNorm2d' or \
       module.__class__.__name__ == 'BatchNorm1d':
        module = PACT_Identity()
        return module
    else:
        for n,m in module.named_children():
            module._modules[n] = _hier_bn_to_identity(m)
        return module

def _hier_dropout_to_identity(module):
    if module.__class__.__name__ == 'Dropout':
        module = PACT_Identity()
        return module
    else:
        for n,m in module.named_children():
            module._modules[n] = _hier_dropout_to_identity(m)
        return module

def _hier_bn_quantizer(module, **kwargs):
    if module.__class__.__name__ == 'BatchNorm2d' or \
        module.__class__.__name__ == 'BatchNorm1d':
        gamma = module.weight.data[:].clone().detach()
        beta = module.bias.data[:].clone().detach()
        sigma = torch.sqrt(module.running_var.data[:] + module.eps).clone().detach()
        mu = module.running_mean.data[:].clone().detach()
        dimensions = 1 if module.__class__.__name__ == 'BatchNorm1d' else 2
        module = PACT_QuantizedBatchNormNd(kappa=gamma/sigma, lamda=beta-gamma/sigma*mu, dimensions=dimensions, **kwargs)
        return module
    else:
        for n,m in module.named_children():
            module._modules[n] = _hier_bn_quantizer(m, **kwargs)
        return module

def _hier_bn_dequantizer(module):
    if module.__class__.__name__ == 'PACT_QuantizedBatchNormNd':
        gamma = module.kappa.data[:].clone().detach().flatten()
        beta = module.lamda.data[:].clone().detach().flatten()
        module = torch.nn.BatchNorm2d(weight=gamma, bias=beta)
        return module
    else:
        for n,m in module.named_children():
            module._modules[n] = _hier_bn_dequantizer(m)
        return module

def _hier_integerizer(module, **kwargs):
    if (module.__class__.__name__ == "PACT_Conv2d" or \
        module.__class__.__name__ == "PACT_Conv1d" or \
        module.__class__.__name__ == "PACT_Linear"):
        module.integerize_weights(**kwargs)
        return module
    elif (module.__class__.__name__ == "PACT_QuantizedBatchNormNd"):
        module = PACT_IntegerBatchNormNd(kappa=module.kappa, lamda=module.lamda, eps_in=module.eps_in, eps_kappa=module.eps_kappa, eps_lamda=module.eps_lamda)
        module.integerize_weights(**kwargs)
    elif (module.__class__.__name__ == "PACT_Act"):
        module = PACT_IntegerAct(precision=module.precision, eps_in=module.eps_in, eps_out=module.eps_static, alpha=module.alpha_static, **kwargs)
        module.set_output_eps(**kwargs)
    elif (module.__class__.__name__ == "PACT_IntegerAdd"):
        module.integerized = True
    elif (module.__class__.__name__ == "AvgPool2d"):
        module = PACT_IntegerAvgPool2d(module.kernel_size, stride=module.stride, padding=module.padding, ceil_mode=module.ceil_mode,
            count_include_pad=module.count_include_pad)
    else:
        for n,m in module.named_children():
            module._modules[n] = _hier_integerizer(m, **kwargs)
    return module

def _hier_thresholdizer_pact(module):
    if module.__class__.__name__ == 'PACT_Act':
        module = PACT_ThresholdAct(precision=module.precision, alpha=module.alpha.data[:])
        return module
    else:
        for n,m in module.named_children():
            module._modules[n] = _hier_thresholdizer_pact(m)
        return module

def _hier_quantizer_pact(module, graph=None, **kwargs):
    if module.__class__.__name__ == 'Conv2d':
        W = module.weight.data
        try:
            b = module.bias.data
        except AttributeError:
            b = None
        module = PACT_Conv2d(
            module.in_channels,
            module.out_channels,
            _single(module.kernel_size),
            stride=_single(module.stride),
            padding=_single(module.padding),
            dilation=_single(module.dilation),
            groups=module.groups,
            bias=True if module.bias is not None else False
        )
        module.weight.data = W.clone()
        if b is not None:
            module.bias.data = b.clone()
        return module
    if module.__class__.__name__ == 'Conv1d':
        W = module.weight.data
        try:
            b = module.bias.data
        except AttributeError:
            b = None
        module = PACT_Conv1d(
            module.in_channels,
            module.out_channels,
            _single(module.kernel_size),
            stride=_single(module.stride),
            padding=_single(module.padding),
            dilation=_single(module.dilation),
            groups=module.groups,
            bias=True if module.bias is not None else False
        )
        module.weight.data = W.clone()
        if b is not None:
            module.bias.data = b.clone()
        return module
    if module.__class__.__name__ == 'Linear':
        W = module.weight.data
        try:
            b = module.bias.data
        except AttributeError:
            b = None
        module = PACT_Linear(
            module.in_features,
            module.out_features,
            bias=True if module.bias is not None else False
        )
        module.weight.data = W.clone()
        if b is not None:
            module.bias.data = b.clone()
        return module
    elif module.__class__.__name__ == 'ReLU6':
        module = PACT_Act(alpha=6., **kwargs)
        return module
    elif module.__class__.__name__ == 'ReLU':
        module = PACT_Act(**kwargs)
        return module
    elif module.__class__.__name__ == 'LeakyReLU':
        module = PACT_Act(leaky=module.negative_slope, **kwargs)
        return module
    else:
        for n,m in module.named_children():
            module._modules[n] = _hier_quantizer_pact(m, **kwargs)
        return module

def _hier_dequantizer_pact(module):
    if module.__class__.__name__ == 'PACT_Conv2d':
        W = module.weight.data
        try:
            b = module.bias.data
        except AttributeError:
            b = None
        module = torch.nn.Conv2d(
            module.in_channels,
            module.out_channels,
            _single(module.kernel_size),
            stride=_single(module.stride),
            padding=_single(module.padding),
            dilation=_single(module.dilation),
            groups=module.groups,
            bias=True if module.bias is not None else False
        )
        module.weight.data = W.clone()
        if b is not None:
            module.bias.data = b.clone()
        return module
    if module.__class__.__name__ == 'PACT_Linear':
        W = module.weight.data
        try:
            b = module.bias.data
        except AttributeError:
            b = None
        module = torch.nn.Linear(
            module.in_features,
            module.out_features,
            bias=True if module.bias is not None else False
        )
        module.weight.data = W.clone()
        if b is not None:
            module.bias.data = b.clone()
        return module
    elif module.__class__.__name__ == 'PACT_Act':
        module = torch.nn.ReLU()
        return module
    else:
        for n,m in module.named_children():
            module._modules[n] = _hier_dequantizer_pact(m)
        return module

def _hier_flat_dict_build(module, name):
    for n,m in module.named_children():
        if n == name:
            return m
        elif n == name.split('.')[0]:
            return _hier_flat_dict_build(m, '.'.join(name.split('.')[1:]))
    return module

def integerize_pact(module, eps_in, **kwargs):
    # r"""Takes a PyTorch module in q-deploy stage and makes it integerized, recursively.

    # :param eps_in: input quantum :math:`\varepsilon_{in}`.
    # :type  eps_in: :py:class:`torch.Tensor`
    # :return: output quantum :math:`\varepsilon_{out}`.
    # :rtype:  :py:class:`torch.Tensor`

    # """
    try:
        net = module.module
    except AttributeError:
        net = module
    # assert(hasattr(net, 'graph'))
    # assert(net.graph is not None)
    net.set_eps_in(eps_in)
    net = _hier_integerizer(net, **kwargs)
    net.graph.rebuild_module_dict()
    # if hasattr(module, 'model'):
    #     module.model = net
    # else:
    #     module = net
    return module

def dequantize_pact(module):
    module = _hier_dequantizer_pact(module)
    if hasattr(module, 'graph'):
        if module.graph is not None:
            module.graph.rebuild_module_dict()
    return module

def thresholdize_pact(module, act_dict):
    module = _hier_thresholdizer_pact(module)
    module.fold_thresholds(act_dict)
    if hasattr(module, 'graph'):
        if module.graph is not None:
            module.graph.rebuild_module_dict()
    return module

def bn_to_identity(module):
    module = _hier_bn_to_identity(module)
    if hasattr(module, 'graph'):
        if module.graph is not None:
            module.graph.rebuild_module_dict()
    return module

def dropout_to_identity(module):
    module = _hier_dropout_to_identity(module)
    if hasattr(module, 'graph'):
        if module.graph is not None:
            module.graph.rebuild_module_dict()
    return module

def bn_quantizer(module, **kwargs):
    module = _hier_bn_quantizer(module, **kwargs)
    if hasattr(module, 'graph'):
        if module.graph is not None:
            module.graph.rebuild_module_dict()
    return module

def bn_dequantizer(module):
    module = _hier_bn_dequantizer(module)
    if hasattr(module, 'graph'):
        if module.graph is not None:
            module.graph.rebuild_module_dict()
    return module



#
# bn.py
# Francesco Conti <fconti@iis.ee.ethz.ch>
#
# Copyright (C) 2018-2020 ETH Zurich
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from nemo.precision import Precision
from nemo.quant.pact import *
from nemo.graph import DeployGraph
from torch.nn.modules.utils import _single,_pair
from collections import OrderedDict
import types
import logging
import numpy as np
import copy
import math
import torchvision.models
import re
from nemo.transf.common import *

BN_PRECISION_MAX = 20

def _absorb_affine_bn(self):
    for n,m in self.named_modules():
        if m.__class__.__name__ == "BatchNorm2d":
            m.weight.data[:] = m.weight.data[:] / torch.sqrt(m.running_var.data[:] + m.eps)
            m.bias.data[:] = m.bias.data[:] - m.weight.data[:] * m.running_mean.data[:]
            m.running_var.data[:] = (1. - m.eps)**2
            m.running_mean.data[:] = 0.

def _prime_affine_bn(self):
    for n,m in self.named_modules():
        if m.__class__.__name__ == "BatchNorm2d":
            m.weight.data[:] = m.weight.data[:] * torch.sqrt(m.running_var.data[:] + m.eps)
            m.bias.data[:] = m.bias.data[:] + m.weight.data[:] * m.running_mean.data[:] / torch.sqrt(m.running_var.data[:] + m.eps)

def _freeze_bn(self, reset_stats=False, disable_grad=False):
    r"""Sets :py:class:torch.nn.BatchNorm2d` layers to not collect statistics and keep the current `running_var` and `running_mean`.
    
    """

    for n,m in self.named_modules():
        if m.__class__.__name__ == "BatchNorm2d":
            if reset_stats:
                try:
                    eps = m.eps
                except AttributeError:
                    eps = 0.
                gamma = m.weight.data[:].clone().detach().cpu()
                beta  = m.bias.data[:].clone().detach().cpu()
                sigma = torch.sqrt(m.running_var.data[:] + eps).clone().detach().cpu()
                mu    = m.running_mean.data[:].clone().detach().cpu()
                kappa = gamma/sigma
                lamda = beta-gamma/sigma * mu
                m.weight.data[:]  = kappa
                m.bias.data[:]    = lamda
                m.running_var[:]  = 1.
                m.running_mean[:] = 0.
            m.track_running_stats = True
            m.eval()
            if disable_grad:
                m.weight.requires_grad = False
                m.bias.requires_grad = False

def _calibrate_bn_pact(self, calib_dict={}, kappa_bit_default=14, lamda_bit_default=20, kappa_dict=None, lamda_dict=None, range_factor=8, minmax=False, **kwargs):
    r"""Calibrates BN layer quantization for :py:class:`nemo.quant.pact.PACT_QuantizedBatchNormNd` layers.
    Using BN min-max statistics previously calculated, this method calibrates the number of bits used in BN parameters
    so that both the BN output and the `kappa`, `lamda` affine parameters are representable.
    
    :param kappa_bit_default: Default maximum number of bits for BN multiplicative parameter (can be overridden by `kappa_dict`); default 16.
    :type  kappa_bit_default: int
    :param lamda_bit_default: Default maximum number of bits for BN additive parameter (can be overridden by `lamda_dict`); default 32.
    :type  lamda_bit_default: int
    :param kappa_dict: dictionary of maximum number of bits for `kappa` in specific layers; overrides `kappa_bit_default`. Default None.
    :type  kappa_dict: `dict` or `collections.OrderedDict`
    :param lamda_dict: dictionary of maximum number of bits for `lamda` in specific layers; overrides `lamda_bit_default`. Default None.
    :type  lamda_dict: `dict` or `collections.OrderedDict`
    
    """

    if not calib_dict:
        calib_dict = get_calib_dict_from_supernodes(self)

    module_dict = {}
    for n,m in self.named_modules():
        if (m.__class__.__name__ == "PACT_Act" or \
            m.__class__.__name__ == "PACT_QuantizedBatchNormNd"):
            module_dict[n] = m

    for n,m in self.named_modules():
        if m.__class__.__name__ == "PACT_QuantizedBatchNormNd":
            kappa_max = m.kappa.abs().max()
            lamda_max = m.lamda.abs().max()
            if minmax:
                out_range = max(m.max.abs(), m.min.abs())
            else: # if we do not have minmax statistics, approximate with alpha mult by a range_factor
                out_range = module_dict[calib_dict[n]].alpha * range_factor
            if kappa_dict is not None:
                kappa_bit_lim = kappa_dict[n]
            else:
                kappa_bit_lim = kappa_bit_default
            if lamda_dict is not None:
                lamda_bit_lim = lamda_dict[n]
            else:
                lamda_bit_lim = lamda_bit_default
            eps_lim = max(out_range, lamda_max) / (2**(lamda_bit_lim - 1) - 1)
            eps_kappa_lim = eps_lim / m.eps_in
            kappa_bits = min(int(min(torch.log2(1 + 2*kappa_max / eps_kappa_lim).floor(), kappa_bit_lim)), BN_PRECISION_MAX)
            lamda_bits = min(lamda_bit_lim, BN_PRECISION_MAX)
            m.precision_kappa.set_bits(kappa_bits)
            m.precision_lamda.set_bits(lamda_bits)
            # print("%s %d %d" % (n, kappa_bits, lamda_bits))

def _unfreeze_bn(self):
    r"""Sets :py:class:`torch.nn.BatchNorm2d` layers to collect statistics and update `running_var` and `running_mean`.
    
    """

    for n,m in self.named_modules():
        if m.__class__.__name__ == "BatchNorm2d":
            m.train()

def _prune_empty_bn_pact(self, bn_dict={}, threshold=None):
    if not bn_dict:
        bn_dict = get_bn_dict_from_supernodes(self)

    module_dict = {}
    for n,m in self.named_modules():
        if (m.__class__.__name__ == "PACT_Conv2d" or \
            m.__class__.__name__ == "PACT_Conv1d" or \
            m.__class__.__name__ == "PACT_Linear" or \
            m.__class__.__name__ == "BatchNorm2d" or \
            m.__class__.__name__ == "BatchNorm1d" or \
            m.__class__.__name__ == "PACT_QuantizedBatchNormNd"):
            module_dict[n] = m
    for n_before in bn_dict.keys():
        n_after  = bn_dict[n_before]
        m_before = module_dict[n_before]
        m_after  = module_dict[n_after]
        if threshold is None:
            try:
                eps = (m_before.W_alpha + m_before.W_beta) / (2.**m_before.W_precision.get_bits()-1)
            except AttributeError:
                continue
        else:
            eps = threshold
        if m_before.bias is not None:
            continue
        try:
            m_after.kappa.data[m_before.weight.data.flatten(1).abs().max(1)[0] < eps] = 0.
        except AttributeError:
            m_after.weight.data[m_before.weight.data.flatten(1).abs().max(1)[0] < eps] = 0.
            m_after.running_var.data[m_before.weight.data.flatten(1).abs().max(1)[0] < eps] = 1.

def _fold_bn_pact(self, bn_dict={}, bn_inv_dict={}, eps=None, phi_inv=0., reset_alpha=True, remove_bn=False):
    r"""Performs batch-normalization folding following the algorithm presented in
    https://arxiv.org/abs/1905.04166. It performs both normal folding and inverse
    folding using two separate dictionaries `bn_dict` and `bn_inv_dict`.
    
    :param bn_dict: a dictionary of layer names, with the key being the source (linear) and the value the target (batch-norm). If empty (default), uses the graph to fold all BN layers.
    :type  bn_dict: `dict` or `collections.OrderedDict`
    :param bn_inv_dict: a dictionary of layers, with the key being the source (batch-norm) and the value the target (linear).
    :type  bn_inv_dict: `dict` or `collections.OrderedDict`
    :param verbose: if True, prints more information.
    :type  verbose: bool
    :param eps: if not None (the default), overrides numerical `eps` used within batch-norm layer.
    :type  eps: float
    :param phi_inv: parameter added to `gamma` in inverse folding for better numerical stability (default 0).
    :type  phi_inv: float
    :param reset_alpha: if True (default), reset the clipping parameters of weights.
    :type  reset_alpha: bool
    :param remove_bn: if True, replace all BN layers with identity layers to "physically remove" BN from the network (default False).
    :type  remove_bn: bool

    """

    if not bn_dict:
        bn_dict = get_bn_dict_from_supernodes(self)

    module_dict = {}
    for n,m in self.named_modules():
        if (m.__class__.__name__ == "PACT_Conv2d" or \
            m.__class__.__name__ == "PACT_Conv1d" or \
            m.__class__.__name__ == "PACT_Linear" or \
            m.__class__.__name__ == "BatchNorm2d" or \
            m.__class__.__name__ == "BatchNorm1d" or \
            m.__class__.__name__ == "PACT_QuantizedBatchNormNd"):
            module_dict[n] = m
    param = {}
    bn_list = list(bn_dict.values()) + list(bn_inv_dict.keys())
    for n in bn_list:
        m = module_dict[n]
        # count how many time this occures as a value in the direct map (for bilinear functions e.g. add)
        count = list(bn_dict.values()).count(n)
        if eps is None:
            try:
                eps = m.eps
            except AttributeError:
                eps = 0.
        if m.__class__.__name__ == "PACT_QuantizedBatchNormNd":
            gamma = m.kappa.data[:].clone().detach().cpu().flatten()
            beta  = m.lamda.data[:].clone().detach().cpu().flatten()
            sigma = torch.ones_like(gamma).flatten()
            mu    = torch.zeros_like(beta).flatten()
        else:
            gamma = m.weight.data[:].clone().detach().cpu()
            beta  = m.bias.data[:].clone().detach().cpu()
            sigma = torch.sqrt(m.running_var.data[:] + eps).clone().detach().cpu()
            mu    = m.running_mean.data[:].clone().detach().cpu()
        param[n] = {
            'gamma' : gamma,
            'beta'  : beta, 
            'sigma' : sigma,
            'mu'    : mu,   
            'count' : count
        }
    # direct folding (CONV->BN)
    for n in bn_dict.keys():
        n_bn  = bn_dict[n]
        m     = module_dict[n]
        m_bn  = module_dict[n_bn]
        gamma = param[n_bn]['gamma']
        beta  = param[n_bn]['beta']
        mu    = param[n_bn]['mu']
        sigma = param[n_bn]['sigma']
        count = param[n_bn]['count']
        if count > 1:
            beta  = beta/count
            mu    = mu/count

        th_a = (gamma/sigma).to(m.weight.device)
        th_b = (beta-gamma/sigma * mu).to(m.weight.device)

        m.weight.data[:] = m.weight.data[:] * reshape_before(m, th_a)
        try:
            m.bias.data[:] = th_a * m.bias.data[:] + th_b
        except AttributeError:
            m.bias = torch.nn.Parameter(th_b)
    # inverse folding (BN->CONV)
    for n_bn in bn_inv_dict.keys():
        n    = bn_inv_dict[n_bn]
        m    = module_dict[n]
        m_bn = module_dict[n_bn]
        gamma = param[n_bn]['gamma']
        beta  = param[n_bn]['beta']
        mu    = param[n_bn]['mu']
        sigma = param[n_bn]['sigma']
        count = param[n_bn]['count']

        th_a = sigma/gamma
        shape_w = np.prod(np.asarray(m.weight.data.shape)[3:1])
        th_m_by_w   = (reshape_after(m, mu)*m.weight.data[:]).sum(3).sum(2).sum(1) / shape_w
        th_bsg_by_w = (reshape_after(m, beta*sigma/(gamma+phi_inv))*m.weight.data[:]).sum(3).sum(2).sum(1) / shape_w

        if phi_inv is None:
            phi_inv = m_bn.eps
        m.weight.data[:] = m.weight.data[:] * reshape_after(m, th_a)
        try:
            m.bias.data[:] = m.bias.data[:] + th_m_by_w - th_bsg_by_w
        except AttributeError:
            m.bias = torch.nn.Parameter(th_m_by_w - th_bsg_by_w)
    # neutralize BatchNorm's
    for n in bn_list:
        m = module_dict[n]
        if m.__class__.__name__ == "PACT_QuantizedBatchNormNd":
            m.kappa.data[:] = 1.
            m.lamda.data[:] = 0.
        else:
            m.weight.data[:] = 1.
            m.bias.data[:] = 0.
            m.running_mean.data[:] = 0.
            m.running_var.data[:] = (1. - eps)**2
    if reset_alpha:
        self.reset_alpha_weights()

def _threshold_folding_pact(self, act_dict):
    r"""Performs the folding of batch-normalization layers into threshold-based
    activation layers.
    
    :param act_dict: a dictionary of layer names, with the key being the source (batch-norm) and the value the target (activation).
    :type  act_dict: `dict` or `collections.OrderedDict`

    """

    module_dict = {}
            
    for n,m in self.named_modules():
        if (m.__class__.__name__ == "PACT_ThresholdAct" or \
            m.__class__.__name__ == "BatchNorm2d" or \
            m.__class__.__name__ == "BatchNorm1d" ):
            module_dict[n] = m
    for n_before in act_dict.keys():
        n_after = act_dict[n_before]
        m_before = module_dict[n_before]
        m_after  = module_dict[n_after]
        # get BN parameters
        eps = m_before.eps
        gamma = m_before.weight.data[:]
        beta  = m_before.bias.data[:]
        sigma = torch.sqrt(m_before.running_var.data[:] + eps)
        mu    = m_before.running_mean.data[:]
        # setup threshold in PACT_ThresholdAct
        del m_after.kappa, m_after.lamda
        m_after.kappa = torch.nn.Parameter(torch.zeros(gamma.shape[0]).to(m_after.alpha.data.device))
        m_after.lamda = torch.nn.Parameter(torch.zeros(gamma.shape[0]).to(m_after.alpha.data.device))
        m_after.kappa.data[:] = sigma/gamma
        m_after.lamda.data[:] = mu - beta*sigma/gamma
        # remove BN parameters
        m_before.weight.data[:] = 1.
        m_before.bias.data[:] = 0.
        m_before.running_var.data[:] = (1. - eps)**2
        m_before.running_mean.data[:] = 0.
